/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include "lik.h"
#include "fF84.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

static double mu[MAX_CATEGORIES], kappa_plus1[MAX_CATEGORIES];
static double freqR, freqY, freqAG, freqCT, freqA2, freqC2, freqG2, freqT2, F84temp1, F84temp2;
static double tab1A, tab2A, tab3A;
static double tab1C, tab2C, tab3C;
static double tab1G, tab2G, tab3G;
static double tab1T, tab2T, tab3T;
double tsTvRatio[MAX_CATEGORIES];
int estimateFreqs;

/*static prototypes*/
static void PrintUsage();
static void PrintVerbose();


/*************************************/
static void PrintUsage()
{
	fprintf(stderr, "  model selected = F84\n");
	fprintf(stderr, "  parameters available: [-f #A #C #G #T|-f=] [-t #]\n");
	fprintf(stderr, "    -f: #A #C #G #T = frequences of bases [default = estimate]\n");
	fprintf(stderr, "    -f=: set frequences of bases equal\n");
	fprintf(stderr, "    -t: # = transition/transversion ratio [default = estimate]\n");
}


/*************************************/
static void PrintVerbose()
{
	fprintf(stderr, "Felsenstein's Phylip DNAML model (F84)\n");
	if (globalMask[0])
		fprintf(stderr, "  estimating ML transition/transversion ratio\n");
	else 
		fprintf(stderr, "  transition/transversion ratio = %f (parameter=%f)\n", tsTvRatio[0], globalMLE[0]);

	if (estimateFreqs)
		fprintf(stderr, "  estimating base frequences from data\n");
	else
		fprintf(stderr, "  base frequences supplied by user\n");
	fprintf(stderr, "    frequencies = A:%f C:%f G:%f T:%f\n", freq[A], freq[C], freq[G], freq[T]);
}


/*************************************/
void Init_Params_F84(int printHelp)
{
	int cat;

	if (printHelp) {
		PrintUsage();
		return;
	}
	
/* The first global MLE's is the Ts/Tv parameter, kappa */	
	estimateGlobals=1;
	for (cat=0; cat<MAX_CATEGORIES; cat++) {
		globalMask[0]=1;
		tsTvRatio[cat]=2.0;
	}

	estimateFreqs=1;
	freq[0]=-999;
	freq[1]=-999;
	freq[2]=-999;
	freq[3]=-999;
}


/*************************************/
int Read_Param_F84(int argc, char **argv, int *index)
{
	char ch;
	char *P;
	
	P=argv[*index];
	P++;
	ch=*P;
	P++;
	switch (ch) {
		case 'F':
			if (*P=='=') {
				freq[0]=0.25;
				freq[1]=0.25;
				freq[2]=0.25;
				freq[3]=0.25;
			} else {
				if (GetDoubleParams(argc, argv, index, P, 4, freq)) {
					fprintf(stderr, "Bad Base Frequencies: %s\n\n", argv[*index]);
					return -1;
				}
			}
			estimateFreqs=0;
		break;
		case 'T':
			if (GetDoubleParams(argc, argv, index, P, 1, tsTvRatio)) {
				fprintf(stderr, "Bad Ts/Tv ratio: %s\n\n", argv[*index]);
				return -1;
			}
			estimateGlobals=0;
			globalMask[0]=0;
		break;
		default:
			fprintf(stderr, "Unknown option: %s\n\n", argv[*index]);
			return -1;
		break;
	}

	return 0;
}


/*************************************/
int Finish_Params_F84()
{
	return 0;
}


/*************************************/
void Init_F84(double *globals, TNode *nodes)
{
	int i, cat;
	double xi, xv;

	numGlobalMLE=FF84_NUM_GLOBALS*numModelCats;

	SetBaseFreq(nodes);
	freqR=freq[A]+freq[G];
	freqY=freq[C]+freq[T];
	freqAG=freq[A]*freq[G];
	freqCT=freq[C]*freq[T];
	freqA2=freq[A]*freq[A];
	freqC2=freq[C]*freq[C];
	freqG2=freq[G]*freq[G];
	freqT2=freq[T]*freq[T];
	F84temp1=1-(freqA2+freqC2+freqG2+freqT2);
	F84temp2=1-((freqA2/freqR)+(freqC2/freqY)+(freqG2/freqR)+(freqT2/freqY));

	tab1A=freqA*((1/freqR)-1);
	tab2A=(freqR-freqA)/freqR;
	tab3A=freqA/freqR;
	tab1C=freqC*((1/freqY)-1);
	tab2C=(freqY-freqC)/freqY;
	tab3C=freqC/freqY;
	tab1G=freqG*((1/freqR)-1);
	tab2G=(freqR-freqG)/freqR;
	tab3G=freqG/freqR;
	tab1T=freqT*((1/freqY)-1);
	tab2T=(freqY-freqT)/freqY;
	tab3T=freqT/freqY;
	
	for (cat=0; cat<numModelCats; cat++) {
		xi=freqR*freqY*(freqR*freqY*tsTvRatio[cat]-freqAG-freqCT);	
		xv=(freqCT*freqR)+(freqAG*freqY);
		globalMLE[cat]=xi/xv;
		Global_F84(cat);
	}
	
	if (!quiet && verbose)
		PrintVerbose();
}


/*************************************/
void Global_F84(int globalIdx)
{
	mu[globalIdx]=-1/(F84temp1+(globalMLE[globalIdx]*F84temp2));
	kappa_plus1[globalIdx]=globalMLE[globalIdx]+1;
}


/*************************************/
void Label_F84(void)
{
	int cat;
	
	if (!quiet && globalMask[0]) {
		if (numModelCats>1)
			for (cat=0; cat<numModelCats; cat++)
				fprintf(stderr, "\tts/tv%d", cat+1);
		else
			fprintf(stderr, "\tts/tv");
	}
}


/*************************************/
void Done_F84(TNode *nodes)
{
	int cat;
	
	for (cat=0; cat<numModelCats; cat++)
		tsTvRatio[cat]=(freqY*(freqAG*freqR+freqCT*freqR)+globalMLE[cat]*(freqCT*freqR+freqAG*freqY))/(freqR*freqR*freqY*freqY);
	
	if (!quiet && globalMask[0]) {
		if (numModelCats>1) {
			for (cat=0; cat<numModelCats; cat++) {
				if (numSets>1) 
					fprintf(stderr, "\t%f", tsTvRatio[cat]);
				else
					fprintf(stderr, "Estimated Ts/Tv for site category %d: %f (K=%f)\n", cat+1, tsTvRatio[cat], globalMLE[cat]);
			}
		} else {
			if (numSets>1) 
				fprintf(stderr, "\t%f", tsTvRatio[0]);
			else
				fprintf(stderr, "Estimated value of Ts/Tv: %f (kappa=%f)\n", tsTvRatio[0], globalMLE[0]);
		}
	}
}


/*********************/
void SetBaseFreq(TNode *nodes)
{
	int i, n, k, weight;
	double total, temp, tempfreq[4], x, difference, sumFreq;

	if (estimateFreqs) {
		for (i=A; i<=T; i++)
			freq[i]=0.25;
		do {
			for(i=A; i<=T; i++)
				tempfreq[i]=0.0;
			total=0.0;
			for(n=0; n<numTaxa; n++) {
				for (k=0; k<numPatterns; k++) {
					weight=patternWeights[0][k];
					for(i=1; i<numCategories; i++) 
						weight+=patternWeights[i][k];
					
					temp=0.0;
					for(i=A; i<=T; i++)
						temp+=nodes[n].Likelihood[k*4+i]*freq[i];
					for(i=A; i<=T; i++) {
						x=(nodes[n].Likelihood[k*4+i]*(freq[i]/temp))*weight;
						tempfreq[i]+=x;
						total+=x;
					}
				}
			}
			difference=0.0;
			for (i=A; i<=T; i++) {
				difference+=fabs((tempfreq[i]/total)-freq[i]);
				freq[i]=tempfreq[i]/total;
			}
		} while (difference>FTOL);
	} else {		
		sumFreq=freq[A]+freq[C]+freq[G]+freq[T];
		if (sumFreq!=1.0) {
			freq[A]*=1.0/sumFreq;
			freq[C]*=1.0/sumFreq;
			freq[G]*=1.0/sumFreq;
			freq[T]*=1.0/sumFreq;
		}
	}
	freqA=freq[0];
	freqC=freq[1];
	freqG=freq[2];
	freqT=freq[3];
}


#define PIJ_SAME_A freqA+(tab1A*aa)+(tab2A*bb)
#define PIJ_TS_A freqA+(tab1A*aa)-(tab3A*bb)
#define PIJ_TV_A freqA*(1-aa)

#define PIJ_SAME_C freqC+(tab1C*aa)+(tab2C*bb)
#define PIJ_TS_C freqC+(tab1C*aa)-(tab3C*bb)
#define PIJ_TV_C freqC*(1-aa)
	
#define PIJ_SAME_G freqG+(tab1G*aa)+(tab2G*bb)
#define PIJ_TS_G freqG+(tab1G*aa)-(tab3G*bb)
#define PIJ_TV_G freqG*(1-aa)
	
#define PIJ_SAME_T freqT+(tab1T*aa)+(tab2T*bb)
#define PIJ_TS_T freqT+(tab1T*aa)-(tab3T*bb)
#define PIJ_TV_T freqT*(1-aa)	
	

/*************************************/
void SetMatrix_F84(double *matrix, int cat, int branchNo)
{
	double xx, aa, bb;
	
	if (numModelCats>1) {
		xx=siteRate[cat]*mu[cat]*branchValue[branchNo];
		bb=exp(kappa_plus1[cat]*xx);
	} else {
		xx=siteRate[cat]*mu[0]*branchValue[branchNo];
		bb=exp(kappa_plus1[0]*xx);
	}
	aa=exp(xx);
	
	matrix[0]=PIJ_SAME_A;
	matrix[1]=PIJ_TV_C;
	matrix[2]=PIJ_TS_G;
	matrix[3]=PIJ_TV_T;

	matrix[4]=PIJ_TV_A;
	matrix[5]=PIJ_SAME_C;
	matrix[6]=PIJ_TV_G;
	matrix[7]=PIJ_TS_T;
	
	matrix[8]=PIJ_TS_A;
	matrix[9]=matrix[1];  /* PIJ_TV_C */
	matrix[10]=PIJ_SAME_G;
	matrix[11]=matrix[3]; /* PIJ_TV_T */
	
	matrix[12]=matrix[4]; /* PIJ_TV_A */
	matrix[13]=PIJ_TS_C;
	matrix[14]=matrix[6]; /* PIJ_TV_G */
	matrix[15]=PIJ_SAME_T;
	
}


/*************************************/
void SetAllMatrices_F84(double *matrix, int cat, int useMask)
{
	int i;
	double *P, *Q;
	char *R;
	double xx, yy, kp1, aa, bb;
	
	P=matrix;
	Q=branchValue;
	if (numModelCats>1) {
		xx=siteRate[cat]*mu[cat];
		kp1=kappa_plus1[cat];
	} else {
		xx=siteRate[cat]*mu[0];
		kp1=kappa_plus1[0];
	}

	R=branchChanged;
	for (i=0; i<numBranches; i++) {
		if (!useMask || *R) {
			yy=xx*(*Q);
			aa=exp(yy);
			bb=exp(yy*kp1);
			
			P[0]=PIJ_SAME_A;
			P[1]=PIJ_TV_C;
			P[2]=PIJ_TS_G;
			P[3]=PIJ_TV_T;

			P[4]=PIJ_TV_A;
			P[5]=PIJ_SAME_C;
			P[6]=PIJ_TV_G;
			P[7]=PIJ_TS_T;
			
			P[8]=PIJ_TS_A;
			P[9]=P[1];  /* PIJ_TV_C */
			P[10]=PIJ_SAME_G;
			P[11]=P[3]; /* PIJ_TV_T */
			
			P[12]=P[4]; /* PIJ_TV_A */
			P[13]=PIJ_TS_C;
			P[14]=P[6]; /* PIJ_TV_G */
			P[15]=PIJ_SAME_T;
		}
		R++;
		Q++;
		P+=16;
	}
}
