/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include "lik.h"
#include "fF84.h"
#include "fHKY.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

static double beta[MAX_CATEGORIES], A_R[MAX_CATEGORIES], A_Y[MAX_CATEGORIES];
static double freqR, freqY;
static double tab1A, tab2A, tab3A;
static double tab1C, tab2C, tab3C;
static double tab1G, tab2G, tab3G;
static double tab1T, tab2T, tab3T;

/*static prototypes*/
static void PrintUsage();
static void PrintVerbose();


/*************************************/
static void PrintUsage()
{
	fprintf(stderr, "  model selected = HKY85\n");
	fprintf(stderr, "  parameters available: [-f #A #C #G #T|-f=] [-t #]\n");
	fprintf(stderr, "    -f: #A #C #G #T = frequences of bases [default = estimate]\n");
	fprintf(stderr, "    -f=: set frequences of bases equal\n");
	fprintf(stderr, "    -t: # = transition/transversion ratio [default = estimate]\n");
}


/*************************************/
static void PrintVerbose()
{
	fprintf(stderr, "HKY85 model\n");
	if (globalMask[0])
		fprintf(stderr, "  estimating ML transition/transversion ratio\n");
	else 
		fprintf(stderr, "  transition/transversion ratio = %f (kappa=%f)\n", tsTvRatio[0], globalMLE[0]);

	if (estimateFreqs)
		fprintf(stderr, "  estimating base frequences from data\n");
	else
		fprintf(stderr, "  base frequences supplied by user\n");
	fprintf(stderr, "    frequencies = A:%f C:%f G:%f T:%f\n", freq[A], freq[C], freq[G], freq[T]);
}


/*************************************/
void Init_Params_HKY(int printHelp)
{
	int cat;
	
	if (printHelp) {
		PrintUsage();
		return;
	}
	
/* The first global MLE's is the Ts/Tv parameter, kappa */	
	estimateGlobals=1;
	for (cat=0; cat<MAX_CATEGORIES; cat++) {
		tsTvRatio[cat]=2.0;
		globalMask[cat]=1;
	}
	
	estimateFreqs=1;
	freq[0]=-999;
	freq[1]=-999;
	freq[2]=-999;
	freq[3]=-999;
}


/*************************************/
int Read_Param_HKY(int argc, char **argv, int *index)
{
	return Read_Param_F84(argc, argv, index);
}	



/*************************************/
int Finish_Params_HKY()
{
	return 0;
}

/*************************************/
void Init_HKY(double *globals, TNode *nodes)
{
	int cat;
	
	numGlobalMLE=FHKY_NUM_GLOBALS*numModelCats;

	SetBaseFreq(nodes);
	freqR=freqA+freqG;
	freqY=freqC+freqT;
	
	tab1A=freqA*((1/freqR)-1);
	tab2A=(freqR-freqA)/freqR;
	tab3A=freqA/freqR;
	tab1C=freqC*((1/freqY)-1);
	tab2C=(freqY-freqC)/freqY;
	tab3C=freqC/freqY;
	tab1G=freqG*((1/freqR)-1);
	tab2G=(freqR-freqG)/freqR;
	tab3G=freqG/freqR;
	tab1T=freqT*((1/freqY)-1);
	tab2T=(freqY-freqT)/freqY;
	tab3T=freqT/freqY;
			
	for (cat=0; cat<numModelCats; cat++) {
		globalMLE[cat]=(tsTvRatio[cat]*freqR*freqY)/(freqA*freqG+freqC*freqT);
		Global_HKY(cat);
	}
	
	if (!quiet && verbose)
		PrintVerbose();
}


/*************************************/
void Global_HKY(int globalIdx)
{
	beta[globalIdx]=-1/(2*(freqR*freqY + globalMLE[globalIdx]*(freqA*freqG + freqC*freqT)));

	A_R[globalIdx]=1+freqR*(globalMLE[globalIdx]-1);
	A_Y[globalIdx]=1+freqY*(globalMLE[globalIdx]-1);
}


/*************************************/
void Label_HKY(void)
{
	Label_F84();
}


/*************************************/
void Done_HKY(TNode *nodes)
{
	int cat;
	
	for (cat=0; cat<numModelCats; cat++)
		tsTvRatio[cat]=(globalMLE[cat]*(freqA*freqG + freqC*freqT))/(freqR*freqY);
	
	if (!quiet && globalMask[0]) {
		if (numModelCats>1) {
			for (cat=0; cat<numModelCats; cat++) {
				if (numSets>1) 
					fprintf(stderr, "\t%f", tsTvRatio[cat]);
				else
					fprintf(stderr, "Estimated Ts/Tv for site category %d: %f (kappa=%f)\n", cat+1, tsTvRatio[cat], globalMLE[cat]);
			}
		} else {
			if (numSets>1) 
				fprintf(stderr, "\t%f", tsTvRatio[0]);
			else
				fprintf(stderr, "Estimated value of Ts/Tv: %f (kappa=%f)\n", tsTvRatio[0], globalMLE[0]);
		}
	}
}


#define PIJ_SAME_A \
	freqA+(tab1A*aa)+(tab2A*bbR)
#define PIJ_TS_A \
	freqA+(tab1A*aa)-(tab3A*bbR)
#define PIJ_TV_A \
	freqA*(1-aa)

#define PIJ_SAME_C \
	freqC+(tab1C*aa)+(tab2C*bbY)
#define PIJ_TS_C \
	freqC+(tab1C*aa)-(tab3C*bbY)
#define PIJ_TV_C \
	freqC*(1-aa)
	
#define PIJ_SAME_G \
	freqG+(tab1G*aa)+(tab2G*bbR)
#define PIJ_TS_G \
	freqG+(tab1G*aa)-(tab3G*bbR)
#define PIJ_TV_G \
	freqG*(1-aa)
	
#define PIJ_SAME_T \
	freqT+(tab1T*aa)+(tab2T*bbY)
#define PIJ_TS_T \
	freqT+(tab1T*aa)-(tab3T*bbY)
#define PIJ_TV_T \
	freqT*(1-aa)	
	
/*************************************/
void SetMatrix_HKY(double *matrix, int cat, int branchNo)
{
	double xx, aa, bbR, bbY;
	
	if (numModelCats>1) {
		xx=siteRate[cat]*beta[cat]*branchValue[branchNo];
		bbR=exp(xx*A_R[cat]);
		bbY=exp(xx*A_Y[cat]);
	} else {
		xx=siteRate[cat]*beta[0]*branchValue[branchNo];
		bbR=exp(xx*A_R[0]);
		bbY=exp(xx*A_Y[0]);
	}
	aa=exp(xx);
	
	matrix[0]=PIJ_SAME_A;
	matrix[1]=PIJ_TV_C;
	matrix[2]=PIJ_TS_G;
	matrix[3]=PIJ_TV_T;

	matrix[4]=PIJ_TV_A;
	matrix[5]=PIJ_SAME_C;
	matrix[6]=PIJ_TV_G;
	matrix[7]=PIJ_TS_T;
	
	matrix[8]=PIJ_TS_A;
	matrix[9]=matrix[1];  /* PIJ_TV_C */
	matrix[10]=PIJ_SAME_G;
	matrix[11]=matrix[3]; /* PIJ_TV_T */
	
	matrix[12]=matrix[4]; /* PIJ_TV_A */
	matrix[13]=PIJ_TS_C;
	matrix[14]=matrix[6]; /* PIJ_TV_G */
	matrix[15]=PIJ_SAME_T;
	
}


/*************************************/
void SetAllMatrices_HKY(double *matrix, int cat, int useMask)
{
	int i;
	double *P, *Q;
	char *R;
	double xx, yy, aa, aaR, aaY, bbR, bbY;
	
	P=matrix;
	Q=branchValue;
	if (numModelCats>1) {
		xx=siteRate[cat]*beta[cat];
		aaR=A_R[cat];
		aaY=A_Y[cat];
	} else {
		xx=siteRate[cat]*beta[0];
		aaR=A_R[0];
		aaY=A_Y[0];
	}
	
	R=branchChanged;
	for (i=0; i<numBranches; i++) {
		if (!useMask || *R) {
			yy=xx*(*Q);
			aa=exp(yy);
			bbR=exp(yy*aaR);
			bbY=exp(yy*aaY);
			
			P[0]=PIJ_SAME_A;
			P[1]=PIJ_TV_C;
			P[2]=PIJ_TS_G;
			P[3]=PIJ_TV_T;

			P[4]=PIJ_TV_A;
			P[5]=PIJ_SAME_C;
			P[6]=PIJ_TV_G;
			P[7]=PIJ_TS_T;
			
			P[8]=PIJ_TS_A;
			P[9]=P[1];  /* PIJ_TV_C */
			P[10]=PIJ_SAME_G;
			P[11]=P[3]; /* PIJ_TV_T */
			
			P[12]=P[4]; /* PIJ_TV_A */
			P[13]=PIJ_TS_C;
			P[14]=P[6]; /* PIJ_TV_G */
			P[15]=PIJ_SAME_T;
		}
		R++;
		Q++;
		P+=16;
	}
}




