/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lik.h"
#include "tree.h"
#include "fHKY.h"
#include "fF84.h"
#include "fREV.h"


/* used to look up transition type */
/* 0 - stays the same, 1 - transition and 2 - transversion */
/* In order A, C, G, T */
short transType[16]={0, 2, 1, 2, 2, 0, 2, 1, 1, 2, 0, 2, 2, 1, 2, 0};

int numTaxa, numBases, numTrees, numDatasets, numSets, numPatterns, ambiguousBases;
short *patternIndex;
short *patternWeights[MAX_CATEGORIES];
double *patternLik[MAX_CATEGORIES];

char **names;
char **sequences;
char **patterns;

double globalMLE[MAX_GLOBAL_MLE*10];
int globalMask[MAX_GLOBAL_MLE*10];
int numGlobalMLE, estimateGlobals;

double *transMatrix[MAX_CATEGORIES];
double *branchValue;
char *branchChanged;
int numBranches, numSplits;

int numCategories, gammaRates, estimateAlpha, codonRates, numModelCats, estimateRates;
int codonCats[3];
double freqRate[MAX_CATEGORIES], siteRate[MAX_CATEGORIES];
double alpha;

int molClock, datedTips, estimateAbsRate;
double absRate, intervalLimits;
int estimateClocks, estimateClockPos, numClocks;
int estimateRoot, rootTaxon;
double clockRate[MAX_CLOCK_RATES];

int userTrees, userLengths, writeTrees, numLoops, estimateIntervals;
int verbose, verboseMemory, verboseLengths, verboseSiteLik, showProgress, quiet;
int debug, debugSplits, debugTrees;
long totalMem;
double freq[4], freqA, freqC, freqG, freqT;

double bestAbsRate, *bestClockRate;
TNode *bestSplit;

void (*InitFunc)(double *, TNode *);
void (*GlobalFunc)(int);
void (*LabelFunc)(void);
void (*DoneFunc)(TNode *);
void (*MatrixFunc)(double *, int, int);
void (*AllMatricesFunc)(double *, int, int);

/* prototypes */


/* functions */

/*************************************/
void *AllocMem(long n, char *name, char *func, int showInfo)
{
	void *P;
	
	if ( (P=malloc(n))==NULL ) {
		fprintf(stderr, "Out of memory allocating '%s': %s()\n", name, func);
		exit(0);
	}
	
	totalMem+=n;

	if (showInfo && verboseMemory)
		fprintf(stderr, "%s in %s() - %d bytes\n", name, func, n);
	
	return P;
}


/*************************************/
void *CAllocMem(long n, char *name, char *func, int showInfo)
{
	void *P;
	
	if ( (P=calloc(n, 1))==NULL ) {
		fprintf(stderr, "Out of memory allocating '%s': %s()\n", name, func);
		exit(0);
	}
	
	totalMem+=n;
	if (showInfo && verboseMemory)
		fprintf(stderr, "%s in %s() - %d bytes\n", name, func, n);
	
	return P;
}


/*************************************/
int FindTipName(char *aName)
{
	int i;
	
	i=0;
	while (i<numTaxa && strncmp(aName, names[i], MAX_NAME_LEN)!=0) 
		i++;
		
	if (i==numTaxa)
		i=-1;
		
	return i;
}


/*************************************/
void AllocateMemory()
{
	int i;
	long mem;
	
/* firstly the names and sequences */
	mem=totalMem;
	names=(char **)AllocMem(sizeof(char *)*numTaxa, "names", "AllocateMemory", 0);
	sequences=(char **)AllocMem(sizeof(char *)*numTaxa, "sequences", "AllocateMemory", 0);
	for (i=0; i<numTaxa; i++) {
		names[i]=(char *)AllocMem(sizeof(char)*(MAX_NAME_LEN+1), "names[]", "AllocateMemory", 0);
		sequences[i]=(char *)AllocMem(sizeof(char)*numBases, "sequences[]", "AllocateMemory", 0);
	}
 		
/* then the pattern stuff */
	patternIndex=(short *)AllocMem(sizeof(short)*numBases, "patternIndex", "AllocateMemory", 1);
	for (i=0; i<numCategories; i++)
		patternWeights[i]=(short *)CAllocMem(sizeof(short)*numBases, "patternWeights[]", "AllocateMemory", 1);

	/* theoretically if each site has a unique pattern then we need 
	numBases+1 pattern slots, the last being a temporary one. */
	patterns=(char **)AllocMem(sizeof(char *)*(numBases+1), "patterns", "AllocateMemory", 0);
	mem=totalMem;
	for (i=0; i<numBases+1; i++)
		patterns[i]=(char *)AllocMem(sizeof(char)*(numTaxa+1), "patterns[]", "AllocateMemory", 0);
	if (verboseMemory)
		fprintf(stderr, "patterns in AllocateMemory- %d bytes\n", totalMem-mem);

	for (i=0; i<numCategories; i++)
		patternLik[i]=NULL;

/* next the branch lengths and transition matrices */
	numBranches=((numTaxa*2)-3);
	numSplits=numTaxa-1;
	branchValue=(double *)AllocMem(sizeof(double)*(numBranches+1), "branchValue", "AllocateMemory", 1);
	branchChanged=(char *)AllocMem(sizeof(char)*(numBranches+1), "branchChanged", "AllocateMemory", 1);
	for (i=0; i<numCategories; i++)
		transMatrix[i]=(double *)AllocMem(sizeof(double)*16*numBranches, "transMatrix[]", "AllocateMemory", 1);	

/* finally the best split and clock rate stuff (used by maxlik module) */
	bestSplit=(TNode *)AllocMem(sizeof(TNode)*numSplits, "bestSplit", "AllocateMemory", 1);
	bestClockRate=(double *)AllocMem(sizeof(double)*numClocks, "bestClockRate", "AllocateMemory", 1);
}


/*************************************/
void FreeMemory()
{
	int i;
	
	for (i=0; i<numTaxa; i++) {
		free(names[i]);
		free(sequences[i]);
	}
	free(names);
	free(sequences);
	
	free(patternIndex);
	for (i=0; i<numCategories; i++)
		free(patternWeights[i]);

	for (i=0; i<numBases+1; i++)
		free(patterns[i]);
	free(patterns);
	for (i=0; i<numCategories; i++)
		if (patternLik[i]!=NULL)
			free(patternLik[i]);
	
	free(branchValue);
	free(branchChanged);
	for (i=0; i<numCategories; i++)
		free(transMatrix[i]);	
	free(bestSplit);
	free(bestClockRate);
}		


/*************************************/
void CreatePatterns()
{
	int i, j, k, cat;
	char *P;
	
	for (i=0; i<numCategories; i++)
		memset(patternWeights[i], 0, sizeof(short)*numBases);
		
	numPatterns=0;
	cat=0;
	for (i=0; i<numBases; i++) {
		if (codonRates)
			cat=codonCats[i % 3];

		if (cat!=-1) {
			P=patterns[numPatterns];
			for (j=0; j<numTaxa; j++) {
				*P=sequences[j][i];
				P++;
			}
			*P='\0';
			
			k=0;
			while ( k<numPatterns && strcmp(patterns[k], patterns[numPatterns])!=0 )
				k++;
			
			if (k==numPatterns) /* i.e. pattern not found - keep it */
				numPatterns++;
				
			patternWeights[cat][k]++;
			patternIndex[i]=k;
		} else
			patternIndex[i]=-1;
	}
}


/*************************************/
void AllocateTreeLikelihoods(TTree *tree)
{
	int i;
	TNode *P;
	
	P=tree->nodes;
	for (i=0; i<tree->numTips; i++) {
		if (P->Likelihood!=NULL)
			free(P->Likelihood);
		P->Likelihood=(double *)CAllocMem(sizeof(double)*4*numPatterns, "P->Likelihood", "ReallocateTreeLikelihoods", 1);
		P++;
	}	
	
	for (i=0; i<tree->numNodes; i++) {
		/* We only need to have one likelihood array and get each of the */
		/* three structures in the node to point to it.  */
		if (P->Likelihood!=NULL)
			free(P->Likelihood);
		P->Likelihood=(double *)CAllocMem(sizeof(double)*(4*numPatterns), "P->Likelihood", "CreateTree", 1);
		(P+1)->Likelihood=P->Likelihood;
		(P+2)->Likelihood=P->Likelihood;
		P+=3;
	}

	for (i=0; i<numCategories; i++) {
		if (patternLik[i]!=NULL)
			free(patternLik[i]);
		patternLik[i]=(double *)CAllocMem(numPatterns*sizeof(double), "patternLik", "ReallocateTreeLikelihoods", 1);
	}
}


/*************************************/
double SetBaseLik(char base, double *site, int seqNum)
{
	int b;
					
	for (b=A; b<=T; b++)
		site[b]=0.0;
		
	switch (base) {
		case 'A': site[A]=1.0; break;
		case 'C': site[C]=1.0; break;
		case 'G': site[G]=1.0; break;
		case 'T': site[T]=1.0; break;
		case 'U': site[T]=1.0; break;
		case 'M': 
			site[A]=1.0; 
			site[C]=1.0; 
		break;
		case 'R': 
			site[A]=1.0; 
			site[G]=1.0; 
		break;
		case 'W': 
			site[A]=1.0; 
			site[T]=1.0; 
		break;
		case 'S': 
			site[C]=1.0; 
			site[G]=1.0; 
		break;
		case 'Y': 
			site[C]=1.0; 
			site[T]=1.0; 
		break;
		case 'K': 
			site[G]=1.0; 
			site[T]=1.0; 
		break;
		case 'B': 
			site[C]=1.0; 
			site[G]=1.0; 
			site[T]=1.0; 
		break;
		case 'D': 
			site[A]=1.0; 
			site[G]=1.0; 
			site[T]=1.0; 
		break;
		case 'H': 
			site[A]=1.0; 
			site[C]=1.0; 
			site[T]=1.0; 
		break;
		case 'V': 
			site[A]=1.0; 
			site[C]=1.0; 
			site[G]=1.0; 
		break;
		case 'N':
		case 'X':
		case '?':
		case 'O':
		case '-': 
			for (b=A; b<=T; b++) 
				site[b]=1.0; 
		break;
		default:
			fprintf(stderr, "Illegal base ('%c') in sequence %d\n", base, seqNum);
			exit(0);
		break;
	}
	
	return site[A]+site[C]+site[G]+site[T];
}


/*************************************/
/*void SetTipLikelihoods(TTree *tree)
{
	int i, j, k, sum;
	char *P;
	
	ambiguousBases=0;
	for (j=0; j<numPatterns; j++) {
		k=j*4;
		P=patterns[j];
		for (i=0; i<numTaxa; i++) {
			sum=SetBaseLik(*P, tree->nodes[i].Likelihood+k, i+1);
			if (sum>1.0)
				ambiguousBases=1;
				
			P++;
		}
	}
}*/

/*************************************/
void SetTipLikelihoods(TTree *tree)
{
	int i, j, k, b;
	char *P;
	
	for (j=0; j<numPatterns; j++) {
		k=j*4;
		P=patterns[j];
		for (i=0; i<numTaxa; i++) {
			for (b=A; b<=T; b++)
				tree->nodes[i].Likelihood[k+b]=0.0;

			switch (*P) {
				case 'A': tree->nodes[i].Likelihood[k+A]=1.0; break;
				case 'C': tree->nodes[i].Likelihood[k+C]=1.0; break;
				case 'G': tree->nodes[i].Likelihood[k+G]=1.0; break;
				case 'T': tree->nodes[i].Likelihood[k+T]=1.0; break;
				case 'U': tree->nodes[i].Likelihood[k+T]=1.0; break;
				case 'M': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+C]=1.0; 
				break;
				case 'R': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+G]=1.0; 
				break;
				case 'W': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'S': 
					tree->nodes[i].Likelihood[k+C]=1.0; 
					tree->nodes[i].Likelihood[k+G]=1.0; 
				break;
				case 'Y': 
					tree->nodes[i].Likelihood[k+C]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'K': 
					tree->nodes[i].Likelihood[k+G]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'B': 
					tree->nodes[i].Likelihood[k+C]=1.0; 
					tree->nodes[i].Likelihood[k+G]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'D': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+G]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'H': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+C]=1.0; 
					tree->nodes[i].Likelihood[k+T]=1.0; 
				break;
				case 'V': 
					tree->nodes[i].Likelihood[k+A]=1.0; 
					tree->nodes[i].Likelihood[k+C]=1.0; 
					tree->nodes[i].Likelihood[k+G]=1.0; 
				break;
				case 'N':
				case 'X':
				case '?':
				case 'O':
				case '-': 
					for (b=A; b<=T; b++) 
						tree->nodes[i].Likelihood[k+b]=1.0; 
				break;
				default:
					fprintf(stderr, "Illegal base ('%c') in sequence %d\n", i, *P);
					exit(0);
				break;
			}
			P++;
		}
	}
}


/*************************************/
void SetModel(int printHelp)
{
	switch(model){
		case HKY:
 			Init_Params_HKY(printHelp);
			InitFunc=Init_HKY;
			GlobalFunc=Global_HKY;
			LabelFunc=Label_HKY;
			DoneFunc=Done_HKY;
			MatrixFunc=SetMatrix_HKY;
			AllMatricesFunc=SetAllMatrices_HKY;
		break;
		case F84:
 			Init_Params_F84(printHelp);
			InitFunc=Init_F84;
			GlobalFunc=Global_F84;
			LabelFunc=Label_F84;
			DoneFunc=Done_F84;
			MatrixFunc=SetMatrix_F84;
			AllMatricesFunc=SetAllMatrices_F84;
		break;
		case REV:
 			Init_Params_REV(printHelp);
			InitFunc=Init_REV;
			GlobalFunc=Global_REV;
			LabelFunc=Label_REV;
			DoneFunc=Done_REV;
			MatrixFunc=SetMatrix_REV;
			AllMatricesFunc=SetAllMatrices_REV;
		break;
	}
}

/*************************************/
int ReadModelParam(int argc, char **argv, int *index)
{	
	int result;
	
	switch(model){
		case HKY:
			result=Read_Param_HKY(argc, argv, index);
		break;
		case F84:
			result=Read_Param_F84(argc, argv, index);
		break;
		case REV:
			result=Read_Param_REV(argc, argv, index);
		break;
	}
	
	return result;
}


/*************************************/
int FinishModelParam()
{	
	int result;
	
	switch(model){
		case HKY:
			result=Finish_Params_HKY();
		break;
		case F84:
			result=Finish_Params_F84();
		break;
		case REV:
			result=Finish_Params_REV();
		break;
	}
	
	return result;
}


/*************************************/
int GetDoubleParams(int argc, char **argv, int *argn, char *pos, int numParams, double *params)
{
	int i;
	char *st, buf[256];
	
	i=0;
	strcpy(buf, pos);
	st=strtok(buf, "\t,/");
	do {
		if (st==NULL) {
			(*argn)++;
			if ((*argn)<argc)
				strcpy(buf, argv[*argn]);
			else
				return -1;
			st=strtok(buf, "\t,/");
			if (st==NULL)
				return -1;
		}
		
		if (sscanf(st, "%lf", params+i)!=1)
			return -1;
		i++;
		if (i<numParams)
			st=strtok(NULL, "\t,/");
	} while (i<numParams);

	return 0;
}


int GetIntParams(int argc, char **argv, int *argn, char *pos, int numParams, int *params)
{
	int i;
	char *st, buf[256];
	
	i=0;
	strcpy(buf, pos);
	st=strtok(buf, "\t,/");
	do {
		if (st==NULL) {
			(*argn)++;
			if ((*argn)<argc)
				strcpy(buf, argv[*argn]);
			else
				return -1;
			st=strtok(buf, "\t,/");
			if (st==NULL)
				return -1;
		}
		
		if (sscanf(st, "%d", params+i)!=1)
			return -1;
		i++;
		if (i<numParams)
			st=strtok(NULL, "\t,/");
	} while (i<numParams);

	return 0;
}


int GetStrParam(int argc, char **argv, int *argn, char *pos, char *param, int len)
{
	int i;
	char *st, *P, buf[256];
	
	i=0;
	strcpy(buf, pos);
	st=strtok(buf, "\t,/");
	if (st==NULL) {
		(*argn)++;
		if ((*argn)<argc)
			strcpy(buf, argv[*argn]);
		else 
			return -1;
		st=strtok(buf, "\t,/");
		if (st==NULL)
			return -1;
	}
	strncpy(param, st, len);
	param[len]='\0';
	P=param;
	while (*P) {
		*P=toupper(*P);
		P++;
	}
	
	return 0;
}
