/*  Header file for lik.c                              */

/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#ifndef _LIK_H_
#define _LIK_H_

#define MAX_NAME_LEN 64
#define MAX_GLOBAL_MLE 64
#define MAX_CATEGORIES 32
#define MAX_CLOCK_RATES 5

#define MAX_BRANCH_LENGTH_VAL 100.0
#define MAX_SPLIT_GAP_VAL 100.0
#define MAX_ABS_RATE_VAL 1.0
#define MAX_ALPHA_VAL 10000
#define MAX_GLOBAL_VAL 100.0
#define MAX_CAT_RATE_VAL 10.0
#define MAX_CLOCK_RATE_VAL 10.0

#define EPS 1.0E-5			/* a small number - always useful */
#define BIG 1.0E5			/* a big number - always useful */
#define FTOL 1.0E-4			/* tolerance for brent/golden search */
#define FTOL2 1.0E-6		/* final tolerance for brent/golden search */
#define STOL 1.0E-4			/* tolerence for smoothing */
#define QTOL 1.0E-2			/* tolerence for quick smoothing */

typedef struct TNode TNode;
struct TNode {
	short 	tipNo, 			/* the tip (taxon) number, -1 if not tip */
			nodeNo, 		/* the internal node number, -1 if tip */
			branchNo,		/* the branch number */
			rate;			/* the molecular clock rate number */
			
	double 	age, 			/* the age of the node in branch length */
			date, 			/* the date of tip (if applicable) in users units */
			gap;			/* difference in age between ancestor */
							/*  and the eldest of the descendants */
	
	TNode 	*next,			/* the next node in triplet (if not tip) */
			*anc, 			/* the ancestor node (if rooted, clock tree) */
			*des1, 			/*  and the two descendant nodes */
			*des2,
			*branch;		/* the node at the other end of the branch */
			
	double *branchLength;	/* pointer to the branch length */
	double *Likelihood; 	/* array of partial likelihoods for each pattern */
};

typedef struct TTree TTree;
struct TTree {
	short 	isRooted,		/* whether tree is rooted */
			numTips, 		/* the number of tip */
			numNodes, 		/* the number of internal nodes */
			totalNodes;		/* the total number of node structs = 3*numNodes + numTips */
			
	double	present, 		/* the date of the most recent tip in users units */
			range;			/* the range of the tip dates */
	
	TNode	*nodes;			/* array of node structures (first numTips are the tips) */
	TNode	*start;			/* a starting position for calculating likelihood */
	
	TNode	root;			/* a node struct to hold root info (not used in likelihood) */
	double	rootBranch1,	/* distances either side of the root */
			rootBranch2; 
	TNode	**splits;		/* array of pointers to TNodes which are splits */
	TNode	**branches;		/* array of pointers to TNodes on one side of a branch */

	short	clockSplits[MAX_CLOCK_RATES];	/* splits at which clock rates change */
	short 	numClocks;
		
	double	lnL;			/* the ln likelihood of the tree */
};


enum treeTypes {UnrootedTree, RootedTree};

enum nucleotides {A, C, G, T};

extern int numTaxa, numBases, numTrees, numDatasets, numSets, numPatterns, ambiguousBases;
extern short *patternIndex;
extern short *patternWeights[MAX_CATEGORIES];
extern double *patternLik[MAX_CATEGORIES];

extern char **names;
extern char **sequences;
extern char **patterns;

extern double globalMLE[MAX_GLOBAL_MLE*10];
extern int globalMask[MAX_GLOBAL_MLE*10];
extern int numGlobalMLE, estimateGlobals;

extern double *transMatrix[MAX_CATEGORIES];
extern double *branchValue;
extern char *branchChanged;
extern int numBranches, numSplits;

extern int numCategories, gammaRates, estimateAlpha, codonRates, numModelCats, estimateRates;
extern int codonCats[3];
extern double freqRate[MAX_CATEGORIES], siteRate[MAX_CATEGORIES];
extern double alpha;

extern int molClock, datedTips, estimateAbsRate;
extern double absRate, intervalLimits;
extern int estimateClocks, estimateClockPos, numClocks;
extern int estimateRoot, rootTaxon;
extern double clockRate[MAX_CLOCK_RATES];

extern int userTrees, userLengths, writeTrees, numLoops, estimateIntervals;
extern int verbose, verboseMemory, verboseLengths, verboseSiteLik, showProgress, quiet;
extern int debug, debugSplits, debugTrees;
extern long totalMem;
extern double freq[4], freqA, freqC, freqG, freqT;

extern double bestAbsRate, *bestClockRate;
extern TNode *bestSplit;

extern void (*InitFunc)(double *, TNode *);
extern void (*GlobalFunc)(int);
extern void (*LabelFunc)(void);
extern void (*DoneFunc)(TNode *);
extern void (*MatrixFunc)(double *, int, int);
extern void (*AllMatricesFunc)(double *, int, int);


extern int model;

enum {
	F84,
	HKY,
	REV,
	numModels
};

/* used to look up transition type */
/* 0 - stays the same, 1 - transition and 2 - transversion */
extern short transType[16];

/* prototypes */
int FindTipName(char *aName);
void *AllocMem(long n, char *name, char *func, int showInfo);
void *CAllocMem(long n, char *name, char *func, int showInfo);
void AllocateMemory();
void FreeMemory();
void CreatePatterns();
void AllocateTreeLikelihoods(TTree *tree);
double SetBaseLik(char base, double *site, int seqNum);
void SetTipLikelihoods(TTree *tree);
void SetModel(int printHelp);
int ReadModelParam(int argc, char **argv, int *index);
int FinishModelParam();

int GetDoubleParams(int argc, char **argv, int *argn, char *pos, int numParams, double *params);
int GetIntParams(int argc, char **argv, int *argn, char *pos, int numParams, int *params);
int GetStrParam(int argc, char **argv, int *argn, char *pos, char *param, int len);

#endif /* _LIK_H_ */
