/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lik.h"
#include "tree.h"
#include "calclik.h"
#include "maxlik.h"
#include "gamma.h"
#include "search.h"
#include "progress.h"

static int currBranch, currLUT, currSplit, currGlobal, currRate, currClock;
static TTree *currTree;

/* prototypes */
int ProgressBarLength();
int CalcEstRem(double oldDiff, double newDiff);
double TreeLikInterval(double rate);
double UserTreeLik();
double MaxBranchLengths(double tol);
double MaxRootLik(double tol);
double MaxClockBranchLengths(double tol);
double MaxGlobals(double tol);
void RangeGlobal(int global, double start, double inc, int n);
double MaxCodonRates(double tol);
double MaxClockRate(double tol);
double MaxAbsRate(double tol);
double MaxAlpha(double tol);

/* functions used by search routines, brent or golden */
double BranchChangedLik(double);
double SplitChangedLik(double);
double GlobalChangedLik(double);
double RateChangedLik(double);
double ClockRateChangedLik(double value);
double AbsRateChangedLik(double value);
double AlphaChangedLik(double value);


int ProgressBarLength()
{
	int barLength=0;
	if (!userLengths) {
		if (molClock && !estimateRoot) {
			if (estimateAbsRate)
				barLength++;		
			if (estimateClocks)
				barLength+=(numClocks-1);
			barLength+=numSplits;
		} else
			barLength+=numBranches;
	}

	if (estimateGlobals)
		barLength+=numGlobalMLE;
	if (codonRates && estimateRates)
		barLength+=numCategories-1;
	if (gammaRates && estimateAlpha)
		barLength++;
	
	return barLength;
}

int CalcEstRem(double oldDiff, double newDiff)
{
	double ratio, diff;
	int i=0;
	
	ratio=newDiff/oldDiff;
	diff=newDiff;
	
	while (diff>STOL && i<100) {
		diff*=ratio;
		i++;
	}
	
	return i;
}

/*********************/
double TreeLik(TTree *tree)
/* In this entire function we are working with the -ln(Likelihood)
   so that a smaller value is better. We then return the negative. 
   This is because the NR brent function is a minimiser */
{	
	int i, j, cat, oldQuiet, estRem, count;
	double lastLik, lik, diff, oldDiff;
	
	int flag=0;
	
	startNode=tree->start;
	currTree=tree;
	
	if (codonRates && estimateRates)
		siteRate[0]=siteRate[1]=siteRate[2]=1.0;
		
	if (gammaRates && estimateAlpha)
		alpha=0.5;
		
	if (estimateAbsRate)
		absRate=0.00001;

	if (estimateClocks) {
		for (i=0; i<MAX_CLOCK_RATES; i++)
			clockRate[i]=1.0;
	}

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
	
	if (!quiet && showProgress)		
		InitProgressBar(ProgressBarLength());
	
	if (molClock)
		TreeAgesToGaps(tree);
	
	oldQuiet=quiet;
	if (numLoops>1)
		quiet=1;
			
	do {
		lastLik=lik=BIG;
		i=0;
		if (!quiet && debug) {
			fprintf(stderr, "n   -lnL       diff.   ");
			if (codonRates && estimateRates)
				for (j=0; j<numCategories; j++)
					fprintf(stderr, " cp%d   ", j+1);
			if (gammaRates && estimateAlpha)
				fprintf(stderr, " alpha  ", j+1);
			if (molClock && debugSplits)
				for (j=0; j<numSplits; j++)
					fprintf(stderr, " sp%-3d  ", j+1);
			for (j=0; j<numGlobalMLE; j++)
				if (globalMask[j])
					fprintf(stderr, " gl%d    ", j+1);
			if (estimateClocks)
				for (j=0; j<numClocks; j++)
					fprintf(stderr, " cl%d    ", j+1);
			if (estimateAbsRate)
				fprintf(stderr, " abs");
			fprintf(stderr, "\n");
		}
		count = 0;
		do {
			if (!quiet && showProgress)
				DrawProgressBar();
			
			if (!userLengths) {
				if (molClock) {
					if (estimateRoot)
						lik=MaxRootLik(FTOL);
					else
						lik=MaxClockBranchLengths(FTOL);
				} else			
					lik=MaxBranchLengths(FTOL);		
			} else
				lik=UserTreeLik(tree);
				
			if (estimateClocks && !estimateRoot)
				lik=MaxClockRate(FTOL);
				
			if (estimateAbsRate && !estimateRoot)
				lik=MaxAbsRate(FTOL);
				
			if (estimateGlobals)
				lik=MaxGlobals(FTOL);

			if (codonRates && estimateRates)
				lik=MaxCodonRates(FTOL);
				
			if (gammaRates && estimateAlpha)
				lik=MaxAlpha(FTOL);
				
			oldDiff=diff;				
			diff=lastLik-lik;
			lastLik=lik;
			i++;
			
			if (!quiet && showProgress) {
				if (i>2) {
					estRem=CalcEstRem(oldDiff, diff);
					fprintf(stderr, "%d/%d:  %-10.4f %-8.4f\n", i, i+estRem, lik, diff);
				} else
					fprintf(stderr, "%-10.4f %-8.4f\n", lik, diff);
			}
			
			if (!quiet && debug) {
				if (diff>1.0) 
					fprintf(stderr, "%-3d %-10.4f >1.0    ", i, lik, diff);
				else
					fprintf(stderr, "%-3d %-10.4f %-8.4f", i, lik, diff);
				if (codonRates && estimateRates)
					for (j=0; j<numCategories; j++)
						fprintf(stderr, " %-6.4f", siteRate[j]);
				if (gammaRates && estimateAlpha)
					fprintf(stderr, " %-6.4f", alpha);
				if (molClock && debugSplits)
					for (j=0; j<numSplits; j++)
						fprintf(stderr, " %-7.4f", tree->splits[j]->age);
				for (j=0; j<numGlobalMLE; j++)
					if (globalMask[j])
						fprintf(stderr, " %-7.4f", globalMLE[j]);
				if (estimateClocks)
					for (j=0; j<numClocks; j++)
						fprintf(stderr, " %-7.4f", clockRate[j]);
				if (estimateAbsRate)
					fprintf(stderr, " %-8.6f", absRate);
				if (debugTrees)
					WriteTree(stdout, tree, estimateClocks);
				fprintf(stderr, "\n");
			}
			count++;
		} while (diff>STOL && count<100);
		
		numLoops--;
	} while (numLoops>0);

	if (!quiet && showProgress) {
		DrawProgressBar();
	}
	if (!userLengths) {
		if (molClock) {
			if (estimateRoot)
				lik=MaxRootLik(FTOL2);
			else
				lik=MaxClockBranchLengths(FTOL2);
		} else			
			lik=MaxBranchLengths(FTOL2);		
	}
	if (estimateClocks && !estimateRoot)
		lik=MaxClockRate(FTOL2);
	if (estimateAbsRate && !estimateRoot)
		lik=MaxAbsRate(FTOL2);	
	if (estimateGlobals)
		lik=MaxGlobals(FTOL2);
	if (codonRates && estimateRates)
		lik=MaxCodonRates(FTOL2);
	if (gammaRates && estimateAlpha)
		lik=MaxAlpha(FTOL2);
	if (!quiet && showProgress) {
		fprintf(stderr, "]\n");
	}
	quiet=oldQuiet;

	if (molClock) 
		SetTreeLocals(tree);
		
	tree->lnL=-lik;
				
	return -lik;
}


/*************************************/
void IntervalAbsRate(TTree *tree, double maxRate, double maxLik, double *lowRate, double *highRate,
						double *lowDate, double *highDate)
{
	double likLimit, lik, x, tol;
	int error, oldQuiet;

	oldQuiet=quiet;
	quiet=1;

	currTree=tree;
	
	estimateAbsRate=0;
	error=0;
	
	likLimit=maxLik-intervalLimits;

	tol=maxRate * FTOL;
	
	if (debug)
		fprintf(stderr, "Finding the low rate:\n");
		
	x=maxRate;
	do {
		x/=2;
			
		if (x<=1E-30)
			error=1;

		if (!error)
			lik = TreeLikInterval(x) - likLimit;
	} while (lik>0 && !error);
		
	if (!error)
		*lowRate=riddlerRoot(TreeLikInterval, x, maxRate, likLimit, tol);
	else
		*lowRate=x;
	*lowDate=tree->root.age/(*lowRate);
		
	if (debug)
		fprintf(stderr, "Finding the high rate:\n");

	error=0;
	x=maxRate;
	do {
		x*=2;
		
		if (x>MAX_ABS_RATE_VAL)
			error=1;
			
		if (!error)
			lik = TreeLikInterval(x) - likLimit;
	} while (lik>0 && !error);
	
	if (!error)
		*highRate=riddlerRoot(TreeLikInterval, maxRate, x, likLimit, tol);
	else
		*highRate=x;
	*highDate=tree->root.age/(*highRate);
	
	quiet=oldQuiet;
	
	estimateAbsRate=1;
}

/*********************/
double TreeLikInterval(double rate)
{	
	int i, j, cat, oldQuiet;
	double lastLik, lik, diff;
	
	int flag=0;
	
	startNode=currTree->start;
	
	if (codonRates && estimateRates)
		siteRate[0]=siteRate[1]=siteRate[2]=1.0;
		
	if (gammaRates && estimateAlpha)
		alpha=0.5;
		
	absRate=rate;

	if (estimateClocks) {
		for (i=0; i<MAX_CLOCK_RATES; i++)
			clockRate[i]=1.0;
	}

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
	
	if (!quiet && showProgress)		
		InitProgressBar(ProgressBarLength());
	
	if (molClock)
		TreeAgesToGaps(currTree);
		
	lastLik=lik=BIG;
	i=0;
	do {
		if (!userLengths) {
			if (molClock) {
				if (estimateRoot)
					lik=MaxRootLik(FTOL);
				else
					lik=MaxClockBranchLengths(FTOL);
			} else			
				lik=MaxBranchLengths(FTOL);		
		} else
			lik=UserTreeLik(currTree);
			
		if (estimateClocks && !estimateRoot)
			lik=MaxClockRate(FTOL);
			
		if (estimateGlobals)
			lik=MaxGlobals(FTOL);

		if (codonRates && estimateRates)
			lik=MaxCodonRates(FTOL);
			
		if (gammaRates && estimateAlpha)
			lik=MaxAlpha(FTOL);
			
		diff=lastLik-lik;
		lastLik=lik;
		i++;
		
	} while (diff>STOL);
		
	if (molClock) 
		SetTreeLocals(currTree);

	if (debug)
		fprintf(stderr, "rate: %f\tlik: %f\n", absRate, lik);
		
	return -lik;
}


/*************************************/
double UserTreeLik()
{
	int cat;
	double lik;

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
	
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxBranchLengths(double tol)
{
	int i, cat;
	double lik, dummy;
	
	for (i=0; i<numBranches; i++) {
		currBranch=i;
		currLUT=i << 4;
		lik=brent(0.0, branchValue[i], MAX_BRANCH_LENGTH_VAL, BranchChangedLik, tol, &dummy);
		branchValue[i]=dummy;
		for (cat=0; cat<numCategories; cat++)
			MatrixFunc(transMatrix[cat]+currLUT, cat, currBranch);
		if (!quiet && showProgress)
			ProgressBar();
	}
	
	return lik;
}

/*************************************/
double BranchChangedLik(double value)
{
	int cat;
	double lik;

	branchValue[currBranch]=value;

	for (cat=0; cat<numCategories; cat++)
/* update transition matrix look up table for just that branch */	
		MatrixFunc(transMatrix[cat]+currLUT, cat, currBranch);
		
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxRootLik(double tol)
{
	int i, j, maxRoot, oldShowProgress;
	double lik, maxLik;

	oldShowProgress=showProgress;
	showProgress=0;
	maxLik=1.0E30;
	for (i=0; i<numBranches; i++) {
		RerootTree(currTree->branches[i], currTree);
		lik=MaxClockBranchLengths(tol);
		if (estimateClocks)
			MaxClockRate(FTOL);
		if (estimateAbsRate)
			MaxAbsRate(FTOL);
				
		if (lik<maxLik) {
			maxLik=lik;
			maxRoot=i;
				
			/* store best parameters */
			for (j=0; j<numSplits; j++) {
				bestSplit[j].gap=currTree->splits[j]->gap;
				bestSplit[j].rate=currTree->splits[j]->rate;
			}
			for (j=1; j<numClocks; j++)
				bestClockRate[j]=clockRate[j];
			bestAbsRate=absRate;
			
		}
		if (!quiet && showProgress)
			ProgressBar();
	}
	showProgress=oldShowProgress;
	
	/* restore best parameters */
	absRate=bestAbsRate;
	RerootTree(currTree->branches[maxRoot], currTree);
	for (i=0; i<numSplits; i++) {
		currTree->splits[i]->gap=bestSplit[i].gap;
		currTree->splits[i]->rate=bestSplit[i].rate;
	}
	if (estimateClockPos) {
		for (i=1; i<numClocks; i++)
			clockRate[i]=bestClockRate[i];
	}
	TreeGapsToAges(currTree);
	SetTreeLocals(currTree);
	
	return maxLik;
}


/*************************************/
double MaxClockBranchLengths(double tol)
{
	int i;
	double lik, dummy;
	
	for (i=0; i<numSplits; i++) {
		currSplit=i;
		lik=brent(0.0, currTree->splits[i]->gap, MAX_SPLIT_GAP_VAL, 
					SplitChangedLik, tol, &dummy);
		currTree->splits[i]->gap=dummy;
		if (!quiet && showProgress)
			ProgressBar();
			
		TreeAgesToGaps(currTree);
	}
		
	return lik;
}


/*************************************/
double SplitChangedLik(double value)
{
	int cat;
	double lik;

	currTree->splits[currSplit]->gap=value;
	TreeGapsToAges(currTree);
	SetTreeLocals(currTree);

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxGlobals(double tol)
{
	int i;
	double lik, dummy;
	
	for (i=0; i<numGlobalMLE; i++) {
		if (globalMask[i]) {
			currGlobal=i;
			lik=brent(0.001, globalMLE[i], MAX_GLOBAL_VAL, GlobalChangedLik, tol, &dummy);
			globalMLE[i]=dummy;
			if (!quiet && showProgress)
				ProgressBar();
		}
	}

	return lik;
}

/*************************************/
void RangeGlobal(int global, double start, double inc, int n)
{
	int i;
	double lik, v;
	
	currGlobal=global;
	v=start;
	for (i=0; i<n; i++) {
		lik=GlobalChangedLik(v);
		v+=inc;
		fprintf(stdout, "%f\t%f\n", v, lik);
	}
}

/*************************************/ 
double GlobalChangedLik(double value)
{
	int cat;
	double lik;

	globalMLE[currGlobal]=value;
	GlobalFunc(currGlobal);
			
	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxCodonRates(double tol)
{
	int i;
	double lik;
	
	for (i=1; i<numCategories; i++) {
		currRate=i;
		lik=brent(0.0, siteRate[i], numCategories, RateChangedLik, tol, siteRate+i);
		if (!quiet && showProgress)
			ProgressBar();
	}

	return lik;
}

/*************************************/
double RateChangedLik(double value)
{
	int cat;
	double lik;

	siteRate[currRate]=value;
	if (numCategories==2)
		siteRate[0]=2.0-value;
	else
		siteRate[0]=3.0-siteRate[1]-siteRate[2];
	if (siteRate[0]<0.0001)
		siteRate[0]=0.0001;
	
	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}

/*************************************/
double MaxClockRate(double tol)
{
	int rate, i, maxSplit, maxIncNode;
	double lik, maxLik, maxRate;

	for (rate=1; rate<numClocks; rate++) {
		currClock=rate;
		if (estimateClockPos) {
			maxLik=1.0E30;
			for (i=1; i<numSplits; i++) { /* root is split 0 so start from 1 */
				currTree->clockSplits[rate]=i;
				
				SetTreeClockRates(0, currTree);
				lik=brent(0.0, clockRate[rate], MAX_CLOCK_RATE_VAL, ClockRateChangedLik, tol, 
							clockRate+rate);
				if (lik<maxLik) {
					maxLik=lik;
					maxRate=clockRate[rate];
					maxSplit=i;
					maxIncNode=0;
				}
				
				SetTreeClockRates(1, currTree);
				lik=brent(0.0, clockRate[rate], MAX_CLOCK_RATE_VAL, ClockRateChangedLik, tol, 
							clockRate+rate);
				if (lik<maxLik) {
					maxLik=lik;
					maxRate=clockRate[rate];
					maxSplit=i;
					maxIncNode=1;
				}		
			}
			currTree->clockSplits[rate]=maxSplit;
			clockRate[rate]=maxRate;
			SetTreeClockRates(maxIncNode, currTree);
			SetTreeLocals(currTree);
		} else {
			maxLik=brent(0.0, clockRate[rate], MAX_CLOCK_RATE_VAL, ClockRateChangedLik, tol, clockRate+rate);
		}

		if (!quiet && showProgress)
			ProgressBar();
	}
	
	return maxLik;
}

/*************************************/
double ClockRateChangedLik(double value)
{
	int cat;
	double lik;

 	clockRate[currClock]=value;
	SetTreeLocals(currTree);
	
	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxAbsRate(double tol)
{
	double lik;

	lik=brent(0.0, absRate, MAX_ABS_RATE_VAL, AbsRateChangedLik, tol, &absRate);
	
	if (!quiet && showProgress)
		ProgressBar();

	return lik;
}

/*************************************/
double AbsRateChangedLik(double value)
{
	int cat;
	double lik;

	absRate=value;

	TreeGapsToAges(currTree);
	SetTreeLocals(currTree);

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}


/*************************************/
double MaxAlpha(double tol)
{
	double lik;

	lik=brent(0.0, alpha, MAX_ALPHA_VAL, AlphaChangedLik, tol, &alpha);
	
	if (!quiet && showProgress)
		ProgressBar();

	return lik;
}

/*************************************/
double AlphaChangedLik(double value)
{
	int cat;
	double lik;

	alpha=value;
	DiscreteGamma(freqRate, siteRate, alpha, alpha, numCategories, 0);

	for (cat=0; cat<numCategories; cat++)
		AllMatricesFunc(transMatrix[cat], cat, 0);
		
	lik=CalcLik();

	return -lik;
}


