/*  (c) Copyright 1997, Andrew Rambaut				   */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>

#ifdef __MWERKS__
#include <profiler.h>
#include <console.h>
#endif	

#include "lik.h"
#include "calclik.h"
#include "maxlik.h"
#include "tree.h"
#include "fF84.h"
#include "gamma.h"
#include "search.h"

#define PROGRAM_NAME "TipDate"
#define VERSION_NUMBER "Version 1.1"

char *ModelNames[numModels]={
	"F84",
	"HKY",
	"REV"
};

int model, printHelp, writeFastDNAml, treeFile;
char treeFileName[256];
FILE *tree_fv;

/* prototypes */

static void PrintTitle();
static void PrintUsage();
void PrintVerbose(TTree *tree);
void PrintSiteLik(FILE *fv);
void ReadParams();
void ReadFileParams();
void ReadFile();
int OpenTreeFile();
TTree *ReadNextTree(TTree *tree);
void CloseTreeFile();
void WriteFastDNAmlFile(FILE *fv, int userTree);
void WriteLikFile(double lik);

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Analysis of trees with dated tips - TipDate\n");
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 1997-1999 Andrew Rambaut\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{
	fprintf(stderr, "Usage: %s [-mMODEL] [-k|-k #|+k #] [+s|-s #] [+r|-r #] [-p nnn] [+c|-c #1 #2 [#3]] [-e] [-g #] [-a #] [-n #] [-h|-help] [parameters]\n          [-u[l]] [-v[b][f][m][p][s][w]|-q] [treefile]\n", PROGRAM_NAME); 	//++pac
	fprintf(stderr, "  -m: MODEL = F84, HKY or REV [default=F84]\n");
	fprintf(stderr, "  -k: Constant rate (molecular clock) branch lengths [default=off]\n");
	fprintf(stderr, "  +s: Estimate absolute rate of molecular evolution (requires dated tips) [default=off]\n");
	fprintf(stderr, "    or -s: # = absolute rate of molecular evolution (requires dated tips)\n");
	fprintf(stderr, "  -r: # = root with specified outgroup taxa (requires molecular clock)\n");
	fprintf(stderr, "    or +r: Estimate maximum likelihood root (requires molecular clock)\n");
	fprintf(stderr, "  -p: CODON CATEGORIES = 112, 123, 120, etc. [homogeneity]\n");
	fprintf(stderr, "  -c: #1 #2 #3 = codon category rates [default = estimate rates].\n");
	fprintf(stderr, "  -i: # = limit to use estimating confidence intervals [default = 1.92].\n");
	fprintf(stderr, "  -e: estimate seperate models for each site category [default = 1 model].\n");
	fprintf(stderr, "  -g: # = num categories for gamma rate heterogeneity [default = homogeneity].\n");
	fprintf(stderr, "  -a: # = gamma shape (alpha) for site rate heterogeneity [default = estimate].\n");
	fprintf(stderr, "  -n: # = Number of Datasets [default=1]\n");
	fprintf(stderr, "  -h,-help = Give this help message (with details of selected model)\n");
	fprintf(stderr, "  parameters = Specific to model.  To get details use: %s -mMODEL -help \n", PROGRAM_NAME);
	fprintf(stderr, "  -v: Verbose information\n");
	fprintf(stderr, "    b Print final branch lengths\n");
	fprintf(stderr, "    m Memory usage information\n");
	fprintf(stderr, "    p Show indication of progress\n");
	fprintf(stderr, "    s Write likelihoods for each Site\n");
	fprintf(stderr, "    w Don't Write trees [default=write]\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  treefile: Optional file containing trees in PHYLIP format\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j, estimate, modelPerCat;
	char ch;
	char *P, st[4];
	
	model=F84;
	printHelp=0;
	writeFastDNAml=0;
	treeFile=0;
	numDatasets=1;
	
	userTrees=1;
	userLengths=0;
	writeTrees=1;
	verbose=0;
	verboseMemory=0;
	verboseLengths=0;
	verboseSiteLik=0;
	showProgress=0;
	quiet=0;

	estimateGlobals=1;

	debug=0;
	debugSplits=0;
	debugTrees=0;
	
	numCategories=1;
	numModelCats=1;
	modelPerCat=0;
	codonCats[0]=codonCats[1]=codonCats[2]=0;
	for (i=0; i<MAX_CATEGORIES; i++)
		siteRate[i]=1.0;
	codonRates=0;
	estimateRates=1;
	gammaRates=0;
	alpha=1.0;
	estimateAlpha=1;
	
	
	molClock=0;
	datedTips=0;
	estimateAbsRate=0;
	absRate=0.001;

	estimateRoot=0;
	rootTaxon=-1;

	estimateClocks=0;
	estimateClockPos=0;
	numClocks=1;
	
	estimateIntervals=1;
	intervalLimits=1.92;

	for (i=0; i<MAX_CLOCK_RATES; i++)
		clockRate[i]=1.0;
	
	numLoops=1;
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (P[0]=='-') {
			ch=toupper(P[1]);
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'Q':
					quiet=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P+2, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}

	if (!quiet || printHelp)
 		PrintTitle();
	if (printHelp)
 		PrintUsage();
	SetModel(printHelp);
	if (printHelp)
		exit(0);
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (*P!='-' && *P!='+') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			if (*P=='+')
				estimate=1;
			else 
				estimate=0;
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'M': 	/* although this has already been read, we need to
							   use it up in this run through of the params.    */
					GetStrParam(argc, argv, &i, P, st, 3);
				break;
				case 'H':
				case 'Q':
				break;
				case 'K':
					molClock=1;
				break;
				case 'R':
					estimateRoot=estimate;
					if (!estimate) {
						if (GetIntParams(argc, argv, &i, P, 1, &rootTaxon)) {
							fprintf(stderr, "Bad root taxon number: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						rootTaxon--;
						if (rootTaxon<0) {
							fprintf(stderr, "Number of root taxon must be between 1 and the number of taxa.\n\n");
							PrintUsage();
							exit(0);
						}
					}
				break;
				case 'S':
					datedTips=1;
					estimateAbsRate=estimate;
					molClock=1;
					if (!estimate) {
						if (GetDoubleParams(argc, argv, &i, P, 1, &absRate)) {
							fprintf(stderr, "Bad Absolute Rate: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					}
				break;
				case 'P':
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Codon Categories Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					P=st;
					if (strlen(P)==3) {
						int cp1=0, cp2=0, cp3=0;
						
						for (j=0; j<3; j++) {
							codonCats[j]=*P-'1';
							switch (codonCats[j]) {
								case -1: break;  /* exclude this codon position */
								case 0: cp1=1; break;
								case 1: cp2=1; break;
								case 2: cp3=1; break;
								default:
									fprintf(stderr, "Codon categories should be from 0-3 (with at least one 1): %s\n\n", argv[i]);
									PrintUsage();
									exit(0);
							}
							P++;
						}
						if (!cp1 || (cp3 && !cp2)) {
							fprintf(stderr, "Codon categories should be from 0-3 (with at least one 1): %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						numCategories=cp1+cp2+cp3;
						codonRates=1;
					} else {
						fprintf(stderr, "Must have a category for each codon position: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'C':
					estimateRates=0;
					if (numCategories==3) {
						if (GetDoubleParams(argc, argv, &i, P, 3, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rates: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else if (numCategories==2) {
						if (GetDoubleParams(argc, argv, &i, P, 2, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rate: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else {
						fprintf(stderr, "Must have two or three codon categories set to specify rates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'E':
					modelPerCat=1;
				break;
				case 'G':
					if (GetIntParams(argc, argv, &i, P, 1, &numCategories) || numCategories<1 || numCategories>MAX_CATEGORIES) {
						fprintf(stderr, "Bad number of Gamma rate categories: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					gammaRates=1;
				break;
				case 'A':
					if (gammaRates) {
						if (GetDoubleParams(argc, argv, &i, P, 1, &alpha) || alpha<0.0) {
							fprintf(stderr, "Bad alpha shape for Gamma rate heterogeneity: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						estimateAlpha=0;
					} else {
						fprintf(stderr, "Must specify the number of gamma rate categories before setting alpha: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'U':
					userTrees=1;
					if (*P=='L')
						userLengths=1;
				break;
				case 'I':
					if (GetDoubleParams(argc, argv, &i, P, 1, &intervalLimits)) {
						fprintf(stderr, "Bad confidence interval limit: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					if (intervalLimits == 0.0) 
						estimateIntervals=0;
				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numDatasets) || numDatasets<1) {
						fprintf(stderr, "Bad number of Datasets: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'V':
					verbose=1;
					while (*P!='\0') {
						switch (*P) {
							case 'P': showProgress=1; break;
							case 'M': verboseMemory=1; break;
							case 'B': verboseLengths=1; break;
							case 'S': verboseSiteLik=1; break;
							case 'W': writeTrees=0; break;
							default:					
								fprintf(stderr, "Unknown verbose mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				case 'D':
					debug=1;
					while (*P!='\0') {
						switch (*P) {
							case 'S': debugSplits=1; break;
							case 'T': debugTrees=1; break;
							case 'W': writeTrees=1; break;
							default:					
								fprintf(stderr, "Unknown debug mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				case 'L':
					if (sscanf(P, "%ld", &numLoops)!=1 || numLoops<1) {
						fprintf(stderr, "Bad Loop Number: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				default:
					if (ReadModelParam(argc, argv, &i)==-1) {
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}
	
	if (numCategories<2) {
		estimateRates=0;
		estimateAlpha=0;
		gammaRates=0;
	}	
	
	if (gammaRates)
		estimateRates=0;
		
	if (modelPerCat) {
		if (gammaRates) {
			fprintf(stderr, "You can't have a gamma model for each category (at the moment)\n\n");
			PrintUsage();
			exit(0);
		}
		numModelCats=numCategories;
	}
	
	if (FinishModelParam()<0) {
		PrintUsage();
		exit(0);
	}
	
	if ((rootTaxon!=-1 || estimateRoot) && !molClock) {
		fprintf(stderr, "Rooting options can only be used with the molecular clock\n\n");
		PrintUsage();
		exit(0);
	}
}


void ReadFileParams()
{
	char ch, st[256];
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
	ungetc(ch, stdin);
	fgets(st, 255, stdin);
	if ( sscanf( st, " %d %d", &numTaxa, &numBases)!=2 ) {
		fprintf(stderr, "Unable to read parameters from standard input\n");
		exit(0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numBases) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numBases ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numTaxa && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numTaxa );
}


int OpenTreeFile()
{
	char st[256];
	int n;
		
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
		n=0;
		while (!feof(tree_fv)) {
			if (fgetc(tree_fv)==';')
				n++;
		}
		rewind(tree_fv);
	} else {
		tree_fv=NULL;
		fgets(st, 255, stdin);
		if ( sscanf(st, " %d ", &n)!=1 ) {
			fprintf(stderr, "Tree is missing from end of sequence file\n");
			exit(0);
		}
	}
	
	return n;
}

TTree *ReadNextTree(TTree *tree)
{		
	if (treeFile) {
		ReadTree(tree_fv, tree);
	} else {
		ReadTree(stdin, tree);
	}
	
	return tree;
}

void CloseTreeFile()
{
	if (treeFile)
		fclose(tree_fv);
}

void WriteFastDNAmlFile(FILE *fv, int userTree)
{
	int i, j;
	char *P;
	
	fprintf(fv, " %d %d", numTaxa, numBases);
	if (numCategories>1) 
		fprintf(fv, " C");
	if (codonCats[0]<0 || codonCats[2]<0 || codonCats[2]<0) 
		fprintf(fv, " W");
	if (!estimateFreqs)
		fprintf(fv, " F");
	if (!userTree)
		fprintf(fv, " U");
	fprintf(fv, " T\n");
	
	fprintf(fv, "T %f\n", tsTvRatio);
	if (!estimateFreqs)
		fprintf(fv, "F %f %f %f %f\n", freqA, freqC, freqG, freqT);

	if (numCategories==2)
		fprintf(fv, "C 2 %f %f\n", siteRate[0], siteRate[1]);
	if (numCategories==3)
		fprintf(fv, "C 3 %f %f %f\n", siteRate[0], siteRate[1], siteRate[2]);
	
	if (numCategories>1) {
		fprintf(fv, "CATEGORIES");
		for (i=0; i<numBases; i++)
			fputc('1'+codonCats[i%3], fv);
		fputc('\n', fv);
	}
	if (codonCats[0]<0 || codonCats[2]<0 || codonCats[2]<0) {
		fprintf(fv, "WEIGHTS   ");
		for (i=0; i<numBases; i++) {
			if (codonCats[i%3]<0)
				fputc('0', fv);
			else
				fputc('1', fv);
		}
		fputc('\n', fv);
	}
	for (i=0; i<numTaxa; i++) {
		fprintf(fv, "%-10s", names[i]);
		P=sequences[i];
		for (j=0; j<numBases; j++)
			fputc(*P++, fv);
		fputc('\n', fv);
	}	
	
	if (userTree)		
		fprintf(fv, "1\n");
}


void WriteLikFile(double lik)
{
	FILE *fv;
	
	if ( (fv=fopen("likfile", "wt"))==NULL ) {
		fprintf(stderr, "Error opening likfile\n");
		exit(0);
	}
	fprintf(fv, "%f\n", lik);
	fclose(fv);
}

void PrintVerbose(TTree *tree)
{
	int i;
	
	fprintf(stderr, "%d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "	%d distinct patterns\n", numPatterns);
	if (userTrees)
		if (numTrees==1)
			fprintf(stderr, "Calculating likelihood of user supplied tree,\n");
		else
			fprintf(stderr, "Calculating likelihood of %d user supplied trees,\n", numTrees);
	if (userLengths)
		fprintf(stderr, "  using branch lengths of supplied tree(s).\n");
	else {
		fprintf(stderr, "  iterating to find maximum likelihood branch lengths");
		if (molClock) {
			if (datedTips) {
				fprintf(stderr, ",\n  assuming a molecular clock with dated taxa.\n");
				fprintf(stderr, "    Taxa dates range between %G and %G,\n", 
						tree->present-tree->range, tree->present);
				if (estimateAbsRate)
					fprintf(stderr, "    estimating the absolute rate of molecular evolution.\n");
				else
					fprintf(stderr, "    using the absolute rate supplied = %f.\n", absRate);
			} else
				fprintf(stderr, ",\n  assuming a molecular clock.\n");
			if (rootTaxon!=-1)
				fprintf(stderr, "  Rooting tree using taxon %d (%s).\n", rootTaxon+1, names[rootTaxon]);
			if (estimateRoot)
				fprintf(stderr, "  Estimating maximum likelihood root position.\n");
		} else
			fprintf(stderr, ".\n");
	}
	if (numCategories>1) {
		fprintf(stderr, "Rate heterogeneity:\n");
		if (gammaRates) {
			fprintf(stderr, "  Discrete Gamma rate heterogeneity with %d categories\n", numCategories);
			if (estimateAlpha)
				fprintf(stderr, "  estimating shape parameter, alpha.\n");
			else
				fprintf(stderr, "  shape parameter, alpha = %f\n", alpha);
		} else {
			fprintf(stderr, "  Number of codon position categories = %d, pattern = %d%d%d\n", 
							numCategories, codonCats[0]+1, codonCats[1]+1, codonCats[2]+1);
			if (estimateRates)
				fprintf(stderr, "  estimating ML rates for categories");
			else
				fprintf(stderr, "  rates for categories supplied by user\n");
		}
		if (!estimateAlpha && !estimateRates) {
			fprintf(stderr, "    rates =");
			for (i=0; i<numCategories; i++)
				fprintf(stderr, " %f", siteRate[i]);
			fputc('\n', stderr);
		}
	} else
		fprintf(stderr, "Rate homogeneity assumed");
	fprintf(stderr, ".\n");
	fprintf(stderr, "Model=%s\n", ModelNames[model]);
}
	
	
void PrintSiteLik(FILE *fv)
{
	short *T;
	int k;

	T=patternIndex;
	fprintf(fv, "1");
	for(k=1; k<=numBases; k++) {
		fprintf(fv, "\t%f", patternLik[*T]);
		T++;
		if (k%6==0 && k<numBases) {
			fprintf(fv, "\n%d", k+1);
		}
	}
	fprintf(fv, "\n\n");
}

/*
void EstimateRate(TTree *tree, double maxRate, double maxLik, double *lowRate, double *highRate)
{
	double rate1, step, stop, last;

	estimateAbsRate=0;
	
	last=maxRate;
	step=0.001;
	stop=maxLik-intervalLimits;
	
	if (debug)
		fprintf(stderr, "MaxLik=%f\tStop=%f\n", maxLik, stop);

	do {
		rate1=last;
		if (debug)
			fprintf(stderr, "Step=%f\tRate=%f\n", step, rate1);
		do {
			last=rate1;
			rate1-=step;

			absRate=rate1;

			TreeLikInterval(tree);
			
			if (debug)
				fprintf(stderr, "\t\tRate=%f\tDiff=%f\n", rate1, maxLik-tree->lnL);
		} while (tree->lnL>stop);
		step/=10;
	} while (fabs(tree->lnL-stop)>0.0001);

	*lowRate=rate1;
		
	last=maxRate;
	step=0.001;
	do { 
		rate1=last;
		if (debug)
			fprintf(stderr, "Step=%f\tRate=%f\n", step, rate1);
		do {
			last=rate1;
			rate1+=step;
			
			absRate=rate1;
			TreeLikInterval(tree);
			
			if (debug)
				fprintf(stderr, "\t\tRate=%f\tDiff=%f\n", rate1, maxLik-tree->lnL);
		} while (tree->lnL>=stop);
		step/=10;
	} while (fabs(tree->lnL-stop)>0.0001);
	*highRate=rate1;

	estimateAbsRate=1;
	absRate=maxRate;
}
*/

int main(int argc, char **argv)
{
	int i, n, oldTaxa, oldBases;
	double lik, maxRate, highRate, lowRate, maxDate, highDate, lowDate;
	TTree *tree;
	clock_t coreStart, totalStart;
	double coreSecs, totalSecs;
	
#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif	
	setvbuf( stderr, NULL, _IOLBF, 512 );
	setvbuf( stdout, NULL, _IOLBF, 512 );

	totalMem=0;
 	
	totalStart = clock();
	
	ReadParams(argc, argv);

	ReadFileParams();
	
	AllocateMemory();
		
	ReadFile();
	
	if (gammaRates)
		DiscreteGamma(freqRate, siteRate, alpha, alpha, numCategories, 0);

	CreatePatterns();

	tree=CreateTree(molClock);
	AllocateTreeLikelihoods(tree);
		
	/*if (userTrees)*/
	numTrees=OpenTreeFile();
	if (numDatasets!=numTrees) {
		if (numDatasets>1 && numTrees>1) {
			fprintf(stderr, "The number of trees and number of datasets don't match.\n\n");
			PrintUsage();
			exit(0);
		}
	}
	ReadNextTree(tree);
	if (numTrees>1)
		numSets=numTrees;
	else 
		numSets=numDatasets;
			
	
	if (rootTaxon>=numTaxa) {
		fprintf(stderr, "Number of root taxon must be between 1 and the number of taxa.\n\n");
		PrintUsage();
		exit(0);
	}

	if (!quiet && verbose)
 		PrintVerbose(tree);

	SetTipLikelihoods(tree);
	SetupLikelihood(tree);
	 		
	coreStart = clock();

	if (numSets>1) {
		if (!quiet) {
			fprintf(stderr, "no.\tlnL");
			LabelFunc();
			if (estimateRates) {
				if (numCategories==2)
					fprintf(stderr, "\trate1\trate2");
				else
					fprintf(stderr, "\trate1\trate2\trate3");
			}
		
			if (estimateClocks) {
				for (i=0; i<numClocks; i++) 
					fprintf(stderr, "\tclock%d", i);
			}
			
			if (datedTips && estimateAbsRate) {
				if (estimateIntervals)
					fprintf(stderr, "\tage\thigh\tlow\trate\thigh\tlow");
				else
					fprintf(stderr, "\tage\trate");
			}
			fprintf(stderr, "\n");
		}
		
		for (n=0; n<numSets; n++) {
			TreeLik(tree);
			
			if (!quiet) {
				lik=tree->lnL;
				
				fprintf(stderr, "%d\t%f", n+1, lik);
				
				FinishLikelihood(tree);

				if (estimateRates) {
					if (numCategories==2)
						fprintf(stderr, "\t%f\t%f", siteRate[0], siteRate[1]);
					else
						fprintf(stderr, "\t%f\t%f\t%f", siteRate[0], siteRate[1], siteRate[2]);
				}
			
				if (estimateClocks) {
					for (i=0; i<numClocks; i++) 
						fprintf(stderr, "\t%f", clockRate[i]);
				}
				
				if (writeTrees) {
					WriteTree(stdout, tree, estimateClocks);
					fprintf(stdout, "\n");
				}
				
				if (datedTips && estimateAbsRate) {
					if (estimateIntervals) {
						maxRate=absRate;
						maxDate=tree->root.age/maxRate;
						IntervalAbsRate(tree, absRate, lik, &lowRate, &highRate, &lowDate, &highDate);
						fprintf(stderr, "\t%f\t%f\t%f\t%f\t%f\t%f", 
							maxDate, highDate, lowDate,
							maxRate, highRate, lowRate);
					} else 
						fprintf(stderr, "\t%f\t%f", tree->root.age/absRate, absRate);
				}
				fprintf(stderr, "\n");
			}
			if (verboseSiteLik)
				PrintSiteLik(stdout);
			
			if (n<numDatasets-1) {
				oldTaxa=numTaxa; oldBases=numBases;
				ReadFileParams();
				if (oldTaxa!=numTaxa || oldBases!=numBases) {
					fprintf(stderr, "The number of taxa and bases in each dataset must be the same.\n");
					exit(0);
				}       
				ReadFile();
				CreatePatterns();
				AllocateTreeLikelihoods(tree);
				SetTipLikelihoods(tree);
				SetupLikelihood(tree);
				if (!treeFile) {
					if (OpenTreeFile()!=1) {
						fprintf(stderr, "Tree missing from end of dataset %d.\n", n+1);
						exit(0);
					}       
					ReadNextTree(tree);
				}
			}

			if (n<numTrees-1)
				ReadNextTree(tree);
		}
	} else {
		TreeLik(tree);
	}
	
	if (numSets==1) {
		WriteTree(stdout, tree, estimateClocks);
		fprintf(stdout, "\n");
	}

	if (writeFastDNAml && numTrees==1) {
		WriteFastDNAmlFile(stdout, 1);
		WriteLikFile(tree->lnL);
	}

	if (!quiet) {
		if (numSets==1) {
			FinishLikelihood(tree);
			if (verboseLengths)
				PrintBranchLengths(tree);
			if (gammaRates && estimateAlpha)
				fprintf(stderr, "  Estimated gamma shape parameter, alpha: %f\n", alpha);
			else if (estimateRates) {
				if (numCategories==2)
					fprintf(stderr, "Estimated codon position rates: %f, %f\n", 
									siteRate[0], siteRate[1]);
				else
					fprintf(stderr, "Estimated codon position rates: %f, %f, %f\n", 
									siteRate[0], siteRate[1], siteRate[2]);
			}


			if (estimateClocks) {
				fprintf(stderr, "Estimated clock rates: %f", clockRate[0]);
				for (i=1; i<numClocks; i++) 
					fprintf(stderr, ", %f", clockRate[i]);
				fprintf(stderr, "\n");
			}
			
			if (datedTips) {
				if (estimateAbsRate) {
					maxRate=absRate;
					maxDate=tree->root.age/maxRate;
					fprintf(stderr, "Estimated absolute rate of molecular evolution: %f\n", maxRate);
					fprintf(stderr, "Absolute age of tree: %f (in the same units as the taxon dates,\n", maxDate);
					fprintf(stderr, "\tmeasured to the most recent tip).\n", maxDate);
					if (estimateIntervals) {
						fprintf(stderr, "\nEstimating confidence intervals...\n");
						fflush(stderr);
						lik=tree->lnL;
						IntervalAbsRate(tree, absRate, lik, &lowRate, &highRate, &lowDate, &highDate);
						tree->lnL=lik;
						fprintf(stderr, "Estimated absolute rate of molecular evolution: %f (%f, %f)\n", maxRate, highRate, lowRate);
						fprintf(stderr, "Absolute age of tree: %f (%f, %f) - in the same units as the taxon dates\n", 
							maxDate, highDate, lowDate);
					}
				} else 
					fprintf(stderr, "Absolute age of tree: %f (in the same units as the taxon dates)\n", tree->root.age/absRate);
			}
			
			fprintf(stderr, "\nTree ln Likelihood = %f\n", tree->lnL);
			if (verboseSiteLik) {
				fprintf(stderr, "\nSite Likelihoods:\n");
				PrintSiteLik(stderr);
			}
		}
	}
	
	coreSecs = (double)(clock() - coreStart) / CLOCKS_PER_SEC;		
	
	CloseTreeFile();
	
	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;

	if (!quiet)
		fprintf(stderr, "\nTime taken - core: %G, total: %G seconds\n", coreSecs, totalSecs);
		
	if (!quiet && verboseMemory)
		fprintf(stderr, "Total memory used: %d\n", totalMem);
	
	
}



