/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "lik.h"
#include "tree.h"


static TNode *nextFreeNode;
static int nextBranch, nextSplit;


/* prototypes */

char ReadToNextChar(FILE *fv);
void ReadUntil(FILE *fv, char stopChar, char *what);
TNode *ReadTip(FILE *fv, char ch, TTree *tree);
TNode *ReadNode(FILE *fv, TNode *node, TTree *tree);
void ConnectBranch(TNode *destNode, TNode *srcNode);
TNode *ReadBranch(FILE *fv, TTree *tree);

void SetNodeAge(double age, TNode *node);
void SetTreeAge(TTree *tree);
void SetTipDates(TTree *tree);

int SetSplitNode(int *nextSplit, TNode *node, TTree *tree);
void SetSplits(TTree *tree);

void SetNodeClockRate(short rate, int incNode, TNode *node);
void SetNodeLocals(TNode *node);

double NodeGapsToAges(TNode *node);

void WriteUnrootedNode(FILE *fv, TNode *node);
void WriteRootedNode(FILE *fv, TNode *node, int printRates);


/* functions */

TTree *CreateTree(short isRooted)
{
	TTree *tree;
	TNode *P;
	int i;
	
	tree=(TTree *)AllocMem(sizeof(TTree), "tree", "CreateTree", 1);

	tree->isRooted=isRooted;
	tree->numTips=numTaxa;
	tree->numNodes=numTaxa-2;
	tree->totalNodes=numTaxa+(3*tree->numNodes);
	tree->numClocks=1;

	tree->nodes=(TNode *)AllocMem(sizeof(TNode)*tree->totalNodes, "tree->nodes", "CreateTree", 1);

	P=tree->nodes;
	for (i=0; i<tree->numTips; i++) {
		P->branch=NULL;
		P->next=NULL;
		
		P->Likelihood=NULL;	
		P->tipNo=i;
		P->nodeNo=-1;
		P->rate=0;
		P->age=0.0;
		P->gap=0.0;
		P->date=0.0;
		P->anc=P->des1=P->des2=NULL;
	
		P++;
	}	
	
	tree->start=P;
	
	for (i=0; i<tree->numNodes; i++) {
		
		P->Likelihood=NULL;
		P->branch=NULL;
		P->tipNo=-1;
		P->nodeNo=i;
		P->rate=0;
		P->next=P+1;
		P->age=P->gap=P->date=0.0;
		P->anc=P->des1=P->des2=NULL;
		
		(P+1)->branch=NULL;
		(P+1)->Likelihood=NULL;
		(P+1)->tipNo=-1;
		(P+1)->nodeNo=i;
		(P+1)->rate=0;
		(P+1)->next=P+2;
		(P+1)->age=(P+1)->gap=(P+1)->date=0.0;
		(P+1)->anc=(P+1)->des1=(P+1)->des2=NULL;
		
		(P+2)->branch=NULL;
		(P+2)->Likelihood=NULL;
		(P+2)->tipNo=-1;
		(P+2)->nodeNo=i;
		(P+2)->rate=0;
		(P+2)->next=P;
		(P+2)->age=(P+2)->gap=(P+2)->date=0.0;
		(P+2)->anc=(P+2)->des1=(P+2)->des2=NULL;
		
		P+=3;
	}
	
	tree->branches=(TNode **)AllocMem(sizeof(TNode *)*numBranches, "tree->branches", "CreateTree", 1);
	if (tree->isRooted) {
		tree->splits=(TNode **)AllocMem(sizeof(TNode *)*numSplits, "tree->splits", "CreateTree", 1);
		for (i=0; i<MAX_CLOCK_RATES; i++) 
			tree->clockSplits[i]=-1;
	} else
		tree->splits=NULL;
	
	return tree;
}


void DisposeTree(TTree *tree)
{
	int i;
	
	if (tree!=NULL) {
		for (i=0; i<tree->totalNodes; i++)
			free(tree->nodes[i].Likelihood);
		
		if (tree->nodes!=NULL)
			free(tree->nodes);
		
		if (tree->branches!=NULL)
			free(tree->branches);
			
		if (tree->splits!=NULL)
			free(tree->splits);
		
		free(tree);
	}
}


char ReadToNextChar(FILE *fv)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && isspace(ch)) 
		ch=fgetc(fv);

	return ch;
}


void ReadUntil(FILE *fv, char stopChar, char *what)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=stopChar) 
		ch=fgetc(fv);

	if (feof(fv)) {
		fprintf(stderr, "%s missing in tree file\n", what);
		exit(0);
	}
}


TNode *ReadTip(FILE *fv, char ch, TTree *tree)
{
	int i;
	char st[MAX_NAME_LEN+1];
	TNode *node;
	double tipDate=0.0;
	
	i=0;
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')' && i<MAX_NAME_LEN) { 
		st[i]=ch;
		i++;
		ch=fgetc(fv);
	}
	st[i]='\0';

	if (datedTips) {
		i=strlen(st);
		while (i>0 && !isdigit(st[i]))
			i--;
		
		if (i==0) {
			fprintf(stderr, "No date in tree for taxa: %s\n", st);
			exit(0);
		}
	
		while (i>=0 && (isdigit(st[i]) || st[i]=='.'))
			i--;
			
		i++;
	
		if (sscanf(st+i, "%lf", &tipDate)!=1) {
			fprintf(stderr, "Bad taxa date in tree for taxa: %s\n", st);
			exit(0);
		}
	}
	
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')') 
		ch=fgetc(fv);
		
	if (feof(fv)) {
		fprintf(stderr, "Unexpected end of file for tree\n");
		exit(0);
	}

	ungetc(ch, fv);
	
	if ((i=FindTipName(st))<0) {
		fprintf(stderr, "Unknown taxa lable in tree: %s\n", st);
		exit(0);
	}
	node=(tree->nodes)+i;
	node->age=0.0;
	node->date=tipDate;
	
	return node;
}


TNode *ReadNode(FILE *fv, TNode *node, TTree *tree)
{
	char ch;
	TNode *node2;
	
	node2=ReadBranch(fv, tree);
	ConnectBranch(node->next, node2);
	ReadUntil(fv, ',', "Comma");
	
	node2=ReadBranch(fv, tree);
	ConnectBranch(node->next->next, node2);
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=')' && ch!=',') 
		ch=fgetc(fv);

	if (ch==',') {
		fprintf(stderr, "This program cannot cope with polytomies in the tree.\nPlease resolve them using zero branch lengths.");
		exit(0);
	}

	ch=fgetc(fv);
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')') 
		ch=fgetc(fv);
		
	if (feof(fv)) {
		fprintf(stderr, "Unexpected end of file for tree\n");
		exit(0);
	}

	ungetc(ch, fv);
	
	return node;
}


void ConnectBranch(TNode *destNode, TNode *srcNode)
{
	srcNode->branch=destNode;
	destNode->branchNo=srcNode->branchNo;
	destNode->branchLength=srcNode->branchLength;
	destNode->branch=srcNode;
	destNode->rate=srcNode->rate;
}	


TNode *ReadBranch(FILE *fv, TTree *tree)
{
	char ch;
	double dist;
	int clockNo;
	TNode *node;
	
	ch=ReadToNextChar(fv);
	if (ch=='(') {	/* is a node */
		node=nextFreeNode;
		nextFreeNode+=3;
		node=ReadNode(fv, node, tree);
	} else {		/* is a tip */
		node=ReadTip(fv, ch, tree);
	}
	
	ch=fgetc(fv);
	if (ch==':') {
		if (fscanf(fv, "%lf", &dist)!=1) {
			fprintf(stderr, "Bad branch length in tree.\n");
			exit(0);
		}
	} else {
		dist=0.1;
		ungetc(ch, fv);
	}
	
	ch=ReadToNextChar(fv);
	if (ch=='[') {
		if (fscanf(fv, "%d", &clockNo)==1) {
			if (clockNo>tree->numClocks)
				tree->numClocks=clockNo;
			node->rate=clockNo-1;
		}
		ReadUntil(fv, ']', "Close Square Brackets");
	} else 
		ungetc(ch, fv);
	
/* we must allow an extra branch length in branchValue in case */
/* the tree is rooted. In this case it will be summed with the */
/* branch on the other side of the root */
	node->branchNo=nextBranch;
	branchValue[nextBranch]=dist;
	node->branchLength=branchValue+nextBranch;
	tree->branches[nextBranch]=node;
		
	nextBranch++;
	
	return node;
}	


void ReadTree(FILE *fv, TTree *tree)
{
	char ch;
	TNode *node1, *node2, *node3;
	
	nextFreeNode=tree->nodes+tree->numTips;
	nextBranch=0;
	
	tree->isRooted=molClock;
	
	ReadUntil(fv, '(', "tree");
	node1=ReadBranch(fv, tree);
	ReadUntil(fv, ',', "Comma");
	node2=ReadBranch(fv, tree);
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=',' && ch!=')') 
		ch=fgetc(fv);

	if (ch==',') {		/* the tree is unrooted */
		if (!tree->isRooted || rootTaxon!=-1 || estimateRoot) {
			node3=ReadBranch(fv, tree);
	
			ConnectBranch(nextFreeNode, node1);
			ConnectBranch(nextFreeNode->next, node2);
			ConnectBranch(nextFreeNode->next->next, node3);
		} else {
			fprintf(stderr, "The tree file is not rooted - it must be to assume a molecular clock.\n");
			exit(0);
		}
	} else if (ch==')') {  			/* the tree is rooted - convert it to unrooted */
		if (tree->isRooted) {
			tree->root.des1=node1;
			tree->root.des2=node2;
	
			tree->rootBranch1=(*node1->branchLength);
			tree->rootBranch2=(*node2->branchLength);
		}
		
		(*node1->branchLength)+=*node2->branchLength;
		ConnectBranch(node2, node1);
	} else {
		fprintf(stderr, "Closing bracket missing in tree file.\n");
		exit(0);
	}
	
	ReadUntil(fv, ';', "Semi-colon");

	if (tree->isRooted) {
		tree->root.age=tree->root.date=tree->root.rate=0;
		tree->root.tipNo=tree->root.nodeNo=tree->root.branchNo=-2;
		tree->root.next=tree->root.anc=tree->root.branch=NULL;
		tree->root.branchLength=tree->root.Likelihood=NULL;
		if (rootTaxon!=-1) /* root at specified outgroup taxon */
			RerootTree(tree->nodes+rootTaxon, tree);
		else if (estimateRoot) /* just root at first branch */
			RerootTree(tree->branches[0], tree);
		else {
			SetSplits(tree);
			SetTreeAge(tree);
		}
		if (datedTips)
			SetTipDates(tree);
	}
	
	tree->start=tree->nodes[0].branch;
}


void SetNodeAge(double age, TNode *node)
{
	node->age=age;
	
	if (node->tipNo==-1) {
		SetNodeAge(age+(*node->des1->branchLength), node->des1);
		SetNodeAge(age+(*node->des2->branchLength), node->des2);
	} 
}


void SetTreeAge(TTree *tree)
/* Set the ages for a rooted tree from the branch lengths. */
/* Doesn't require contemporaneous tips tree */
{
	int i;
	double oldest;
	
	SetNodeAge(tree->rootBranch1, tree->root.des1);
	SetNodeAge(tree->rootBranch2, tree->root.des2);
	
	oldest=0.0;
	for (i=0; i<numTaxa; i++) {
		if (tree->nodes[i].age>oldest)
			oldest=tree->nodes[i].age;
	}
	for (i=0; i<numSplits; i++)
		tree->splits[i]->age=oldest-tree->splits[i]->age;
	for (i=0; i<numTaxa; i++)
		tree->nodes[i].age=oldest-tree->nodes[i].age;
}


void SetTipDates(TTree *tree)
{
	int i;
	double maxDate, minDate;
	
/* find the minimum and maximum dates */
	maxDate=minDate=tree->nodes[0].date;
	for (i=1; i<numTaxa; i++) {
		if (tree->nodes[i].date>maxDate)
			maxDate=tree->nodes[i].date;
		if (tree->nodes[i].date<minDate)
			minDate=tree->nodes[i].date;
	}
	
	tree->present=maxDate;
	tree->range=maxDate-minDate;

/* scale the dates between 0 (present) and tree->range */
	for (i=0; i<numTaxa; i++)
		tree->nodes[i].date=maxDate-tree->nodes[i].date;
}


int SetSplitNode(int *nextSplit, TNode *node, TTree *tree)
{
	if (node->tipNo==-1) {
		tree->splits[*nextSplit]=node;
		(*nextSplit)++;
		
		node->des1=node->next->branch;
		node->des1->anc=node;
		SetSplitNode(nextSplit, node->des1, tree);
		
		node->des2=node->next->next->branch;
		node->des2->anc=node;
		SetSplitNode(nextSplit, node->des2, tree);
	}
}



void SetSplits(TTree *tree)
{
	int nextSplit=1;
	
	tree->splits[0]=&tree->root;
	tree->root.des1->anc=&tree->root;
	SetSplitNode(&nextSplit, tree->root.des1, tree);
	tree->root.des2->anc=&tree->root;
	SetSplitNode(&nextSplit, tree->root.des2, tree);

	tree->clockSplits[0]=0;
}


void SetNodeClockRate(short rate, int incNode, TNode *node)
{
	/* set the two descendant nodes */
	node->des1->rate=rate;
	node->des2->rate=rate;
	
	if (incNode)
		node->rate=rate;
	
	if (node->des1->tipNo==-1)
		SetNodeClockRate(rate, 1, node->des1);
		
	if (node->des2->tipNo==-1)
		SetNodeClockRate(rate, 1, node->des2);
}


void SetTreeClockRates(int incNode, TTree *tree)
{
	int i;
	
	SetNodeClockRate(0, 0, &tree->root);
	for (i=1; i<numClocks; i++)
		if (tree->clockSplits[i]!=-1)
			SetNodeClockRate(i, incNode, tree->splits[tree->clockSplits[i]]);
}

void SetNodeLocals(TNode *node)
{
	double dist;
	
	if (node->tipNo==-1) {
		dist=(node->age - node->des1->age)*clockRate[node->des1->rate];
		if (dist<EPS) 			
			dist=EPS;		/* this is a filthy hack */
		if (*(node->des1->branchLength)!=dist) {
			branchChanged[node->des1->branchNo]=1;
		}
		*(node->des1->branchLength)=dist;
		SetNodeLocals(node->des1);
		
		dist=(node->age - node->des2->age)*clockRate[node->des2->rate];
		if (dist<EPS) 			
			dist=EPS;		/* this is also a filthy hack */
		if (*(node->des2->branchLength)!=dist) {
			branchChanged[node->des2->branchNo]=1;
		}
		*(node->des2->branchLength)=dist;
		SetNodeLocals(node->des2);
	}	
}


void SetTreeLocals(TTree *tree)
{
	double dist;
	
	memset(branchChanged, 0, numBranches);
	SetNodeLocals(tree->root.des1);
	SetNodeLocals(tree->root.des2);
	
	dist=((tree->root.age - tree->root.des1->age) * clockRate[tree->root.des1->rate]) + 
		((tree->root.age - tree->root.des2->age) * clockRate[tree->root.des2->rate]);
	if (dist<EPS)
		dist=EPS; 		/* this is a filthy hack as well */
	if (*(tree->root.des1->branchLength)!=dist) {
		branchChanged[tree->root.des1->branchNo]=1;
	}
	*(tree->root.des1->branchLength)=dist;
}


void TreeAgesToGaps(TTree *tree)
{
	int i;
	
	for (i=0; i<numSplits; i++) {
		if (tree->splits[i]->des1->age > tree->splits[i]->des2->age)
			tree->splits[i]->gap=tree->splits[i]->age - tree->splits[i]->des1->age;
		else
			tree->splits[i]->gap=tree->splits[i]->age - tree->splits[i]->des2->age;
		if (tree->splits[i]->gap<0)
			fprintf(stderr, "eek");
	}
}


double NodeGapsToAges(TNode *node)
{
	double d1, d2, diff, newGap;
	
	if (node->tipNo<0) {
		d1=NodeGapsToAges(node->des1);
		d2=NodeGapsToAges(node->des2);
		
		if (d1>d2)
			node->age=node->gap+d1;
		else 
			node->age=node->gap+d2;
	} else
		node->age=node->date*absRate;
		
	return node->age;
}


void TreeGapsToAges(TTree *tree)
{
	NodeGapsToAges(&tree->root);
}


void RerootTree(TNode *node, TTree *tree)
{
	int i;
	double rootLen;
	
	rootLen=(*node->branchLength)/2;
	tree->rootBranch1=rootLen;
	tree->rootBranch2=rootLen;
	tree->root.des1=node;
	tree->root.des2=node->branch;
	
	SetSplits(tree);
	if (datedTips) {
		for (i=0; i<numTaxa; i++) 
			tree->nodes[i].age=0.0;
	}
	SetTreeAge(tree);
}


void WriteUnrootedNode(FILE *fv, TNode *node)
{
	if (node->tipNo<0) {
		fprintf(fv, "(");
		WriteUnrootedNode(fv, node->next->branch);
		fprintf(fv, ":%.8f,", *node->next->branchLength);
		WriteUnrootedNode(fv, node->next->next->branch);
		fprintf(fv, ":%.8f)", *node->next->next->branchLength);
	} else {
		fprintf(fv, "%s", names[node->tipNo]);
	}
}


void WriteRootedNode(FILE *fv, TNode *node, int printRates)
{
	if (node->tipNo<0) {
		fprintf(fv, "(");
		WriteRootedNode(fv, node->next->branch, printRates);
		if (printRates)
			fprintf(fv, ":%.8f[%d],", node->age-node->next->branch->age, 
										node->next->branch->rate+1);
		else
			fprintf(fv, ":%.8f,", node->age-node->next->branch->age);
			
		WriteRootedNode(fv, node->next->next->branch, printRates);
		if (printRates)
			fprintf(fv, ":%.8f[%d])", node->age-node->next->next->branch->age, 
										node->next->next->branch->rate+1);
		else
			fprintf(fv, ":%.8f)", node->age-node->next->next->branch->age);
	} else {
		fprintf(fv, "%s", names[node->tipNo]);
	}
}


void WriteTree(FILE *fv, TTree *tree, int printRates)
{
	if (tree->isRooted) {
		fprintf(fv, "(");
		WriteRootedNode(fv, tree->root.des1, printRates);
		if (printRates)
			fprintf(fv, ":%.8f[%d],", tree->root.age - tree->root.des1->age, 
										tree->root.des1->rate+1);
		else
			fprintf(fv, ":%.8f,", tree->root.age - tree->root.des1->age);
		WriteRootedNode(fv, tree->root.des2, printRates);
		if (printRates)
			fprintf(fv, ":%.8f[%d]);", tree->root.age - tree->root.des2->age, 
										tree->root.des2->rate+1);
		else
		fprintf(fv, ":%.8f);\n", tree->root.age - tree->root.des2->age);
	} else {
		fprintf(fv, "(");
		WriteUnrootedNode(fv, tree->start->next->branch);
		fprintf(fv, ":%.8f,", *tree->start->next->branchLength);
		WriteUnrootedNode(fv, tree->start->next->next->branch);
		fprintf(fv, ":%.8f,", *tree->start->next->next->branchLength);
		WriteUnrootedNode(fv, tree->start->branch);
		fprintf(fv, ":%.8f);\n", *tree->start->branchLength);
	}
}


void PrintBranchLengths(TTree *tree)
{
	TNode *P;
	int i;
	int *branches;
	
	branches=(int *)CAllocMem(sizeof(int)*numBranches, "branches", "PrintBranchLengths", 0);
	
	P=tree->nodes;
	fprintf(stderr, "\nLength\tFrom\tTo\n");
	for (i=0; i<tree->numTips; i++) {
		fprintf(stderr, "%5.4f\t%3hd\t%s\n", *P->branchLength, P->branch->nodeNo+1, names[P->tipNo]);
		branches[P->branchNo]=1;
		P++;
	}	
	for (i=0; i<numBranches; i++) {
		if (!branches[P->branchNo]) {
			fprintf(stderr, "%5.4f  %3hd to %3hd\n", *P->branchLength, P->nodeNo+1, P->branch->nodeNo+1);
			branches[P->branchNo]=1;
		}
		P++;
	}	
	
	free(branches);
}
